/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.model;

import com.fasterxml.jackson.annotation.JsonProperty;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import javax.annotation.Generated;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;

/**
 * Immutable implementation of {@link ModelExport}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutableModelExport.builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableModelExport.of()}.
 */
@SuppressWarnings({"varargs", "all"})
@SuppressFBWarnings
@Generated({"Immutables.generator", "ModelExport"})
final class ImmutableModelExport implements ModelExport {
  private final String schemaVersion;
  private final List<ModelData<?>> models;

  private ImmutableModelExport(String schemaVersion, Iterable<? extends ModelData<?>> models) {
    this.schemaVersion = schemaVersion;
    this.models = createUnmodifiableList(false, createSafeList(models, false, true));
  }

  private ImmutableModelExport(
      ImmutableModelExport original,
      String schemaVersion,
      List<ModelData<?>> models) {
    this.schemaVersion = schemaVersion;
    this.models = models;
  }

  /**
   * @return The value of the {@code schemaVersion} attribute
   */
  @JsonProperty("schemaVersion")
  @Override
  public String schemaVersion() {
    return schemaVersion;
  }

  /**
   * @return The value of the {@code models} attribute
   */
  @JsonProperty("models")
  @Override
  public List<ModelData<?>> models() {
    return models;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ModelExport#schemaVersion() schemaVersion} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for schemaVersion (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableModelExport withSchemaVersion(String value) {
    if (Objects.equals(this.schemaVersion, value)) return this;
    return validate(new ImmutableModelExport(this, value, this.models));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ModelExport#models() models}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  @SafeVarargs
  public final ImmutableModelExport withModels(ModelData<?>... elements) {
    List<ModelData<?>> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), false, true));
    return validate(new ImmutableModelExport(this, this.schemaVersion, newValue));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ModelExport#models() models}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of models elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableModelExport withModels(Iterable<? extends ModelData<?>> elements) {
    if (this.models == elements) return this;
    List<ModelData<?>> newValue = createUnmodifiableList(false, createSafeList(elements, false, true));
    return validate(new ImmutableModelExport(this, this.schemaVersion, newValue));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableModelExport} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableModelExport
        && equalTo((ImmutableModelExport) another);
  }

  private boolean equalTo(ImmutableModelExport another) {
    return Objects.equals(schemaVersion, another.schemaVersion)
        && models.equals(another.models);
  }

  /**
   * Computes a hash code from attributes: {@code schemaVersion}, {@code models}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + Objects.hashCode(schemaVersion);
    h += (h << 5) + models.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code ModelExport} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "ModelExport{"
        + "schemaVersion=" + schemaVersion
        + ", models=" + models
        + "}";
  }

  /**
   * Construct a new immutable {@code ModelExport} instance.
   * @param schemaVersion The value for the {@code schemaVersion} attribute
   * @param models The value for the {@code models} attribute
   * @return An immutable ModelExport instance
   */
  public static ModelExport of(String schemaVersion, List<ModelData<?>> models) {
    return of(schemaVersion, (Iterable<? extends ModelData<?>>) models);
  }

  /**
   * Construct a new immutable {@code ModelExport} instance.
   * @param schemaVersion The value for the {@code schemaVersion} attribute
   * @param models The value for the {@code models} attribute
   * @return An immutable ModelExport instance
   */
  public static ModelExport of(String schemaVersion, Iterable<? extends ModelData<?>> models) {
    return validate(new ImmutableModelExport(schemaVersion, models));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableModelExport validate(ImmutableModelExport instance) {
    Set<ConstraintViolation<ImmutableModelExport>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link ModelExport} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable ModelExport instance
   */
  public static ModelExport copyOf(ModelExport instance) {
    if (instance instanceof ImmutableModelExport) {
      return (ImmutableModelExport) instance;
    }
    return ImmutableModelExport.builder()
        .createFrom(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ModelExport ModelExport}.
   * @return A new ModelExport builder
   */
  public static ImmutableModelExport.Builder builder() {
    return new ImmutableModelExport.Builder();
  }

  /**
   * Builds instances of type {@link ModelExport ModelExport}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static final class Builder {
    private String schemaVersion;
    private List<ModelData<?>> models = new ArrayList<ModelData<?>>();

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code ModelExport} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder createFrom(ModelExport instance) {
      Objects.requireNonNull(instance, "instance");
      String schemaVersionValue = instance.schemaVersion();
      if (schemaVersionValue != null) {
        schemaVersion(schemaVersionValue);
      }
      addAllModels(instance.models());
      return this;
    }

    /**
     * Initializes the value for the {@link ModelExport#schemaVersion() schemaVersion} attribute.
     * @param schemaVersion The value for schemaVersion (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("schemaVersion")
    public final Builder schemaVersion(String schemaVersion) {
      this.schemaVersion = schemaVersion;
      return this;
    }

    /**
     * Adds one element to {@link ModelExport#models() models} list.
     * @param element A models element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addModel(ModelData<?> element) {
      if (element != null) this.models.add(element);
      return this;
    }

    /**
     * Adds elements to {@link ModelExport#models() models} list.
     * @param elements An array of models elements
     * @return {@code this} builder for use in a chained invocation
     */
    @SafeVarargs
    public final Builder addModel(ModelData<?>... elements) {
      for (ModelData<?> element : elements) {
        if (element != null) this.models.add(element);
      }
      return this;
    }

    /**
     * Sets or replaces all elements for {@link ModelExport#models() models} list.
     * @param elements An iterable of models elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("models")
    public final Builder models(Iterable<? extends ModelData<?>> elements) {
      this.models.clear();
      return addAllModels(elements);
    }

    /**
     * Adds elements to {@link ModelExport#models() models} list.
     * @param elements An iterable of models elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addAllModels(Iterable<? extends ModelData<?>> elements) {
      for (ModelData<?> element : elements) {
        if (element != null) this.models.add(element);
      }
      return this;
    }

    /**
     * Builds a new {@link ModelExport ModelExport}.
     * @return An immutable instance of ModelExport
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ModelExport build() {
      return ImmutableModelExport.validate(new ImmutableModelExport(null, schemaVersion, createUnmodifiableList(true, models)));
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<T>();
    } else {
      list = new ArrayList<T>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<T>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }
}
