/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.model;

import com.fasterxml.jackson.annotation.JsonProperty;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import javax.annotation.Generated;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;

/**
 * Immutable implementation of {@link Split}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new Split.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableSplit.of()}.
 */
@SuppressWarnings({"all"})
@SuppressFBWarnings
@Generated({"Immutables.generator", "Split"})
final class ImmutableSplit implements Split {
  private final String language;
  private final String expression;

  private ImmutableSplit(Optional<String> language, Optional<String> expression) {
    this.language = language.orElse(null);
    this.expression = expression.orElse(null);
  }

  private ImmutableSplit(ImmutableSplit original, String language, String expression) {
    this.language = language;
    this.expression = expression;
  }

  /**
   * @return The value of the {@code language} attribute
   */
  @JsonProperty("language")
  @Override
  public Optional<String> getLanguage() {
    return Optional.ofNullable(language);
  }

  /**
   * @return The value of the {@code expression} attribute
   */
  @JsonProperty("expression")
  @Override
  public Optional<String> getExpression() {
    return Optional.ofNullable(expression);
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Split#getLanguage() language} attribute.
   * @param value The value for language
   * @return A modified copy of {@code this} object
   */
  public final ImmutableSplit withLanguage(String value) {
    String newValue = Objects.requireNonNull(value, "language");
    if (Objects.equals(this.language, newValue)) return this;
    return validate(new ImmutableSplit(this, newValue, this.expression));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Split#getLanguage() language} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for language
   * @return A modified copy of {@code this} object
   */
  public final ImmutableSplit withLanguage(Optional<String> optional) {
    String value = optional.orElse(null);
    if (Objects.equals(this.language, value)) return this;
    return validate(new ImmutableSplit(this, value, this.expression));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Split#getExpression() expression} attribute.
   * @param value The value for expression
   * @return A modified copy of {@code this} object
   */
  public final ImmutableSplit withExpression(String value) {
    String newValue = Objects.requireNonNull(value, "expression");
    if (Objects.equals(this.expression, newValue)) return this;
    return validate(new ImmutableSplit(this, this.language, newValue));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Split#getExpression() expression} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for expression
   * @return A modified copy of {@code this} object
   */
  public final ImmutableSplit withExpression(Optional<String> optional) {
    String value = optional.orElse(null);
    if (Objects.equals(this.expression, value)) return this;
    return validate(new ImmutableSplit(this, this.language, value));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableSplit} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableSplit
        && equalTo((ImmutableSplit) another);
  }

  private boolean equalTo(ImmutableSplit another) {
    return Objects.equals(language, another.language)
        && Objects.equals(expression, another.expression);
  }

  /**
   * Computes a hash code from attributes: {@code language}, {@code expression}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + Objects.hashCode(language);
    h += (h << 5) + Objects.hashCode(expression);
    return h;
  }

  /**
   * Prints the immutable value {@code Split} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder("Split{");
    if (language != null) {
      builder.append("language=").append(language);
    }
    if (expression != null) {
      if (builder.length() > 6) builder.append(", ");
      builder.append("expression=").append(expression);
    }
    return builder.append("}").toString();
  }

  /**
   * Construct a new immutable {@code Split} instance.
   * @param language The value for the {@code language} attribute
   * @param expression The value for the {@code expression} attribute
   * @return An immutable Split instance
   */
  public static Split of(Optional<String> language, Optional<String> expression) {
    return validate(new ImmutableSplit(language, expression));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableSplit validate(ImmutableSplit instance) {
    Set<ConstraintViolation<ImmutableSplit>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link Split} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Split instance
   */
  public static Split copyOf(Split instance) {
    if (instance instanceof ImmutableSplit) {
      return (ImmutableSplit) instance;
    }
    return new Split.Builder()
        .createFrom(instance)
        .build();
  }

  /**
   * Builds instances of type {@link Split Split}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static class Builder {
    private String language;
    private String expression;

    /**
     * Creates a builder for {@link Split Split} instances.
     */
    public Builder() {
      if (!(this instanceof Split.Builder)) {
        throw new UnsupportedOperationException("Use: new Split.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code Split} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Split.Builder createFrom(Split instance) {
      Objects.requireNonNull(instance, "instance");
      Optional<String> languageOptional = instance.getLanguage();
      if (languageOptional.isPresent()) {
        language(languageOptional);
      }
      Optional<String> expressionOptional = instance.getExpression();
      if (expressionOptional.isPresent()) {
        expression(expressionOptional);
      }
      return (Split.Builder) this;
    }

    /**
     * Initializes the optional value {@link Split#getLanguage() language} to language.
     * @param language The value for language
     * @return {@code this} builder for chained invocation
     */
    public final Split.Builder language(String language) {
      this.language = Objects.requireNonNull(language, "language");
      return (Split.Builder) this;
    }

    /**
     * Initializes the optional value {@link Split#getLanguage() language} to language.
     * @param language The value for language
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("language")
    public final Split.Builder language(Optional<String> language) {
      this.language = language.orElse(null);
      return (Split.Builder) this;
    }

    /**
     * Initializes the optional value {@link Split#getExpression() expression} to expression.
     * @param expression The value for expression
     * @return {@code this} builder for chained invocation
     */
    public final Split.Builder expression(String expression) {
      this.expression = Objects.requireNonNull(expression, "expression");
      return (Split.Builder) this;
    }

    /**
     * Initializes the optional value {@link Split#getExpression() expression} to expression.
     * @param expression The value for expression
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("expression")
    public final Split.Builder expression(Optional<String> expression) {
      this.expression = expression.orElse(null);
      return (Split.Builder) this;
    }

    /**
     * Builds a new {@link Split Split}.
     * @return An immutable instance of Split
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public Split build() {
      return ImmutableSplit.validate(new ImmutableSplit(null, language, expression));
    }
  }
}
