/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.model.action;

import com.fasterxml.jackson.annotation.JsonProperty;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import io.syndesis.model.WithName;
import io.syndesis.model.WithTags;
import java.io.ObjectStreamException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.NavigableSet;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import javax.annotation.Generated;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;

/**
 * Immutable implementation of {@link ExtensionAction}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new ExtensionAction.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableExtensionAction.of()}.
 */
@SuppressWarnings({"all"})
@SuppressFBWarnings
@Generated({"Immutables.generator", "ExtensionAction"})
final class ImmutableExtensionAction implements ExtensionAction {
  private final String actionType;
  private final String description;
  private final ExtensionDescriptor descriptor;
  private final Action.Pattern pattern;
  private final String name;
  private final SortedSet<String> tags;
  private final String id;

  private ImmutableExtensionAction(
      String actionType,
      String description,
      ExtensionDescriptor descriptor,
      Action.Pattern pattern,
      String name,
      Iterable<String> tags,
      Optional<String> id) {
    this.actionType = Objects.requireNonNull(actionType, "actionType");
    this.description = description;
    this.descriptor = descriptor;
    this.pattern = pattern;
    this.name = name;
    this.tags = createUnmodifiableSortedSet(false, createSafeList(tags, false, true));
    this.id = id.orElse(null);
  }

  private ImmutableExtensionAction(ImmutableExtensionAction.Builder builder) {
    this.description = builder.description;
    this.descriptor = builder.descriptor;
    this.pattern = builder.pattern;
    this.name = builder.name;
    this.tags = createUnmodifiableSortedSet(false, createSafeList(builder.tags, false, false));
    this.id = builder.id;
    this.actionType = builder.actionType != null
        ? builder.actionType
        : Objects.requireNonNull(ExtensionAction.super.getActionType(), "actionType");
  }

  private ImmutableExtensionAction(
      ImmutableExtensionAction original,
      String actionType,
      String description,
      ExtensionDescriptor descriptor,
      Action.Pattern pattern,
      String name,
      SortedSet<String> tags,
      String id) {
    this.actionType = actionType;
    this.description = description;
    this.descriptor = descriptor;
    this.pattern = pattern;
    this.name = name;
    this.tags = tags;
    this.id = id;
  }

  /**
   * @return The value of the {@code actionType} attribute
   */
  @JsonProperty("actionType")
  @Override
  public String getActionType() {
    return actionType;
  }

  /**
   * @return The value of the {@code description} attribute
   */
  @JsonProperty("description")
  @Override
  public String getDescription() {
    return description;
  }

  /**
   * @return The value of the {@code descriptor} attribute
   */
  @JsonProperty("descriptor")
  @Override
  public ExtensionDescriptor getDescriptor() {
    return descriptor;
  }

  /**
   * @return The value of the {@code pattern} attribute
   */
  @JsonProperty("pattern")
  @Override
  public Action.Pattern getPattern() {
    return pattern;
  }

  /**
   * @return The value of the {@code name} attribute
   */
  @JsonProperty("name")
  @Override
  public String getName() {
    return name;
  }

  /**
   * @return The value of the {@code tags} attribute
   */
  @JsonProperty("tags")
  @Override
  public SortedSet<String> getTags() {
    return tags;
  }

  /**
   * @return The value of the {@code id} attribute
   */
  @JsonProperty("id")
  @Override
  public Optional<String> getId() {
    return Optional.ofNullable(id);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ExtensionAction#getActionType() actionType} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for actionType
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtensionAction withActionType(String value) {
    if (this.actionType.equals(value)) return this;
    String newValue = Objects.requireNonNull(value, "actionType");
    return validate(new ImmutableExtensionAction(this, newValue, this.description, this.descriptor, this.pattern, this.name, this.tags, this.id));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ExtensionAction#getDescription() description} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for description (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtensionAction withDescription(String value) {
    if (Objects.equals(this.description, value)) return this;
    return validate(new ImmutableExtensionAction(this, this.actionType, value, this.descriptor, this.pattern, this.name, this.tags, this.id));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ExtensionAction#getDescriptor() descriptor} attribute.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for descriptor (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtensionAction withDescriptor(ExtensionDescriptor value) {
    if (this.descriptor == value) return this;
    return validate(new ImmutableExtensionAction(this, this.actionType, this.description, value, this.pattern, this.name, this.tags, this.id));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ExtensionAction#getPattern() pattern} attribute.
   * A value equality check is used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for pattern (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtensionAction withPattern(Action.Pattern value) {
    if (this.pattern == value) return this;
    return validate(new ImmutableExtensionAction(this, this.actionType, this.description, this.descriptor, value, this.name, this.tags, this.id));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ExtensionAction#getName() name} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for name (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtensionAction withName(String value) {
    if (Objects.equals(this.name, value)) return this;
    return validate(new ImmutableExtensionAction(
        this,
        this.actionType,
        this.description,
        this.descriptor,
        this.pattern,
        value,
        this.tags,
        this.id));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ExtensionAction#getTags() tags}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionAction withTags(String... elements) {
    SortedSet<String> newValue = createUnmodifiableSortedSet(false, createSafeList(Arrays.asList(elements), false, true));
    return validate(new ImmutableExtensionAction(
        this,
        this.actionType,
        this.description,
        this.descriptor,
        this.pattern,
        this.name,
        newValue,
        this.id));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ExtensionAction#getTags() tags}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of tags elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionAction withTags(Iterable<String> elements) {
    if (this.tags == elements) return this;
    SortedSet<String> newValue = createUnmodifiableSortedSet(false, createSafeList(elements, false, true));
    return validate(new ImmutableExtensionAction(
        this,
        this.actionType,
        this.description,
        this.descriptor,
        this.pattern,
        this.name,
        newValue,
        this.id));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link ExtensionAction#getId() id} attribute.
   * @param value The value for id
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionAction withId(String value) {
    String newValue = Objects.requireNonNull(value, "id");
    if (Objects.equals(this.id, newValue)) return this;
    return validate(new ImmutableExtensionAction(
        this,
        this.actionType,
        this.description,
        this.descriptor,
        this.pattern,
        this.name,
        this.tags,
        newValue));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link ExtensionAction#getId() id} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for id
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionAction withId(Optional<String> optional) {
    String value = optional.orElse(null);
    if (Objects.equals(this.id, value)) return this;
    return validate(new ImmutableExtensionAction(
        this,
        this.actionType,
        this.description,
        this.descriptor,
        this.pattern,
        this.name,
        this.tags,
        value));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableExtensionAction} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableExtensionAction
        && equalTo((ImmutableExtensionAction) another);
  }

  private boolean equalTo(ImmutableExtensionAction another) {
    return actionType.equals(another.actionType)
        && Objects.equals(description, another.description)
        && Objects.equals(descriptor, another.descriptor)
        && Objects.equals(pattern, another.pattern)
        && Objects.equals(name, another.name)
        && tags.equals(another.tags)
        && Objects.equals(id, another.id);
  }

  /**
   * Computes a hash code from attributes: {@code actionType}, {@code description}, {@code descriptor}, {@code pattern}, {@code name}, {@code tags}, {@code id}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + actionType.hashCode();
    h += (h << 5) + Objects.hashCode(description);
    h += (h << 5) + Objects.hashCode(descriptor);
    h += (h << 5) + Objects.hashCode(pattern);
    h += (h << 5) + Objects.hashCode(name);
    h += (h << 5) + tags.hashCode();
    h += (h << 5) + Objects.hashCode(id);
    return h;
  }

  /**
   * Prints the immutable value {@code ExtensionAction} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder("ExtensionAction{");
    builder.append("actionType=").append(actionType);
    if (description != null) {
      builder.append(", ");
      builder.append("description=").append(description);
    }
    if (descriptor != null) {
      builder.append(", ");
      builder.append("descriptor=").append(descriptor);
    }
    if (pattern != null) {
      builder.append(", ");
      builder.append("pattern=").append(pattern);
    }
    if (name != null) {
      builder.append(", ");
      builder.append("name=").append(name);
    }
    builder.append(", ");
    builder.append("tags=").append(tags);
    if (id != null) {
      builder.append(", ");
      builder.append("id=").append(id);
    }
    return builder.append("}").toString();
  }

  /**
   * Construct a new immutable {@code ExtensionAction} instance.
   * @param actionType The value for the {@code actionType} attribute
   * @param description The value for the {@code description} attribute
   * @param descriptor The value for the {@code descriptor} attribute
   * @param pattern The value for the {@code pattern} attribute
   * @param name The value for the {@code name} attribute
   * @param tags The value for the {@code tags} attribute
   * @param id The value for the {@code id} attribute
   * @return An immutable ExtensionAction instance
   */
  public static ExtensionAction of(String actionType, String description, ExtensionDescriptor descriptor, Action.Pattern pattern, String name, SortedSet<String> tags, Optional<String> id) {
    return of(actionType, description, descriptor, pattern, name, (Iterable<String>) tags, id);
  }

  /**
   * Construct a new immutable {@code ExtensionAction} instance.
   * @param actionType The value for the {@code actionType} attribute
   * @param description The value for the {@code description} attribute
   * @param descriptor The value for the {@code descriptor} attribute
   * @param pattern The value for the {@code pattern} attribute
   * @param name The value for the {@code name} attribute
   * @param tags The value for the {@code tags} attribute
   * @param id The value for the {@code id} attribute
   * @return An immutable ExtensionAction instance
   */
  public static ExtensionAction of(String actionType, String description, ExtensionDescriptor descriptor, Action.Pattern pattern, String name, Iterable<String> tags, Optional<String> id) {
    return validate(new ImmutableExtensionAction(actionType, description, descriptor, pattern, name, tags, id));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableExtensionAction validate(ImmutableExtensionAction instance) {
    Set<ConstraintViolation<ImmutableExtensionAction>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link ExtensionAction} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable ExtensionAction instance
   */
  public static ExtensionAction copyOf(ExtensionAction instance) {
    if (instance instanceof ImmutableExtensionAction) {
      return (ImmutableExtensionAction) instance;
    }
    return new ExtensionAction.Builder()
        .createFrom(instance)
        .build();
  }

  private Object readResolve() throws ObjectStreamException {
    return validate(this);
  }

  /**
   * Builds instances of type {@link ExtensionAction ExtensionAction}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static class Builder {
    private String actionType;
    private String description;
    private ExtensionDescriptor descriptor;
    private Action.Pattern pattern;
    private String name;
    private List<String> tags = new ArrayList<String>();
    private String id;

    /**
     * Creates a builder for {@link ExtensionAction ExtensionAction} instances.
     */
    public Builder() {
      if (!(this instanceof ExtensionAction.Builder)) {
        throw new UnsupportedOperationException("Use: new ExtensionAction.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.action.ExtensionAction} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExtensionAction.Builder createFrom(ExtensionAction instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ExtensionAction.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithName} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExtensionAction.Builder createFrom(WithName instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ExtensionAction.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithTags} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExtensionAction.Builder createFrom(WithTags instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ExtensionAction.Builder) this;
    }

    private void from(Object object) {
      if (object instanceof ExtensionAction) {
        ExtensionAction instance = (ExtensionAction) object;
        Action.Pattern patternValue = instance.getPattern();
        if (patternValue != null) {
          pattern(patternValue);
        }
        actionType(instance.getActionType());
        String descriptionValue = instance.getDescription();
        if (descriptionValue != null) {
          description(descriptionValue);
        }
        ExtensionDescriptor descriptorValue = instance.getDescriptor();
        if (descriptorValue != null) {
          descriptor(descriptorValue);
        }
        Optional<String> idOptional = instance.getId();
        if (idOptional.isPresent()) {
          id(idOptional);
        }
      }
      if (object instanceof WithName) {
        WithName instance = (WithName) object;
        String nameValue = instance.getName();
        if (nameValue != null) {
          name(nameValue);
        }
      }
      if (object instanceof WithTags) {
        WithTags instance = (WithTags) object;
        addAllTags(instance.getTags());
      }
    }

    /**
     * Initializes the value for the {@link ExtensionAction#getActionType() actionType} attribute.
     * <p><em>If not set, this attribute will have a default value as returned by the initializer of {@link ExtensionAction#getActionType() actionType}.</em>
     * @param actionType The value for actionType 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("actionType")
    public final ExtensionAction.Builder actionType(String actionType) {
      this.actionType = Objects.requireNonNull(actionType, "actionType");
      return (ExtensionAction.Builder) this;
    }

    /**
     * Initializes the value for the {@link ExtensionAction#getDescription() description} attribute.
     * @param description The value for description (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("description")
    public final ExtensionAction.Builder description(String description) {
      this.description = description;
      return (ExtensionAction.Builder) this;
    }

    /**
     * Initializes the value for the {@link ExtensionAction#getDescriptor() descriptor} attribute.
     * @param descriptor The value for descriptor (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("descriptor")
    public final ExtensionAction.Builder descriptor(ExtensionDescriptor descriptor) {
      this.descriptor = descriptor;
      return (ExtensionAction.Builder) this;
    }

    /**
     * Initializes the value for the {@link ExtensionAction#getPattern() pattern} attribute.
     * @param pattern The value for pattern (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("pattern")
    public final ExtensionAction.Builder pattern(Action.Pattern pattern) {
      this.pattern = pattern;
      return (ExtensionAction.Builder) this;
    }

    /**
     * Initializes the value for the {@link ExtensionAction#getName() name} attribute.
     * @param name The value for name (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("name")
    public final ExtensionAction.Builder name(String name) {
      this.name = name;
      return (ExtensionAction.Builder) this;
    }

    /**
     * Adds one element to {@link ExtensionAction#getTags() tags} sortedSet.
     * @param element A tags element
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExtensionAction.Builder addTag(String element) {
      if (element != null) this.tags.add(element);
      return (ExtensionAction.Builder) this;
    }

    /**
     * Adds elements to {@link ExtensionAction#getTags() tags} sortedSet.
     * @param elements An array of tags elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExtensionAction.Builder addTag(String... elements) {
      for (String element : elements) {
        if (element != null) this.tags.add(element);
      }
      return (ExtensionAction.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link ExtensionAction#getTags() tags} sortedSet.
     * @param elements An iterable of tags elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("tags")
    public final ExtensionAction.Builder tags(Iterable<String> elements) {
      this.tags.clear();
      return addAllTags(elements);
    }

    /**
     * Adds elements to {@link ExtensionAction#getTags() tags} sortedSet.
     * @param elements An iterable of tags elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExtensionAction.Builder addAllTags(Iterable<String> elements) {
      for (String element : elements) {
        if (element != null) this.tags.add(element);
      }
      return (ExtensionAction.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExtensionAction#getId() id} to id.
     * @param id The value for id
     * @return {@code this} builder for chained invocation
     */
    public final ExtensionAction.Builder id(String id) {
      this.id = Objects.requireNonNull(id, "id");
      return (ExtensionAction.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExtensionAction#getId() id} to id.
     * @param id The value for id
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("id")
    public final ExtensionAction.Builder id(Optional<String> id) {
      this.id = id.orElse(null);
      return (ExtensionAction.Builder) this;
    }

    /**
     * Builds a new {@link ExtensionAction ExtensionAction}.
     * @return An immutable instance of ExtensionAction
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ExtensionAction build() {
      return ImmutableExtensionAction.validate(new ImmutableExtensionAction(this));
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<T>();
    } else {
      list = new ArrayList<T>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T extends Comparable<T>> NavigableSet<T> createUnmodifiableSortedSet(boolean reverse, List<T> list) {
    TreeSet<T> set = reverse
        ? new TreeSet<T>(Collections.reverseOrder())
        : new TreeSet<T>();
    set.addAll(list);
    return Collections.unmodifiableNavigableSet(set);
  }
}
