/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.model.action;

import com.fasterxml.jackson.annotation.JsonProperty;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import io.syndesis.model.DataShape;
import java.io.ObjectStreamException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import javax.annotation.Generated;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;

/**
 * Immutable implementation of {@link ExtensionDescriptor}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new ExtensionDescriptor.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableExtensionDescriptor.of()}.
 */
@SuppressWarnings({"all"})
@SuppressFBWarnings
@Generated({"Immutables.generator", "ExtensionDescriptor"})
final class ImmutableExtensionDescriptor implements ExtensionDescriptor {
  private final ExtensionAction.Kind kind;
  private final String entrypoint;
  private final DataShape inputDataShape;
  private final DataShape outputDataShape;
  private final List<ActionDescriptor.ActionDescriptorStep> propertyDefinitionSteps;

  private ImmutableExtensionDescriptor(
      ExtensionAction.Kind kind,
      String entrypoint,
      Optional<DataShape> inputDataShape,
      Optional<DataShape> outputDataShape,
      Iterable<? extends ActionDescriptor.ActionDescriptorStep> propertyDefinitionSteps) {
    this.kind = kind;
    this.entrypoint = entrypoint;
    this.inputDataShape = inputDataShape.orElse(null);
    this.outputDataShape = outputDataShape.orElse(null);
    this.propertyDefinitionSteps = createUnmodifiableList(false, createSafeList(propertyDefinitionSteps, true, false));
  }

  private ImmutableExtensionDescriptor(
      ImmutableExtensionDescriptor original,
      ExtensionAction.Kind kind,
      String entrypoint,
      DataShape inputDataShape,
      DataShape outputDataShape,
      List<ActionDescriptor.ActionDescriptorStep> propertyDefinitionSteps) {
    this.kind = kind;
    this.entrypoint = entrypoint;
    this.inputDataShape = inputDataShape;
    this.outputDataShape = outputDataShape;
    this.propertyDefinitionSteps = propertyDefinitionSteps;
  }

  /**
   * @return The value of the {@code kind} attribute
   */
  @JsonProperty("kind")
  @Override
  public ExtensionAction.Kind getKind() {
    return kind;
  }

  /**
   * @return The value of the {@code entrypoint} attribute
   */
  @JsonProperty("entrypoint")
  @Override
  public String getEntrypoint() {
    return entrypoint;
  }

  /**
   * @return The value of the {@code inputDataShape} attribute
   */
  @JsonProperty("inputDataShape")
  @Override
  public Optional<DataShape> getInputDataShape() {
    return Optional.ofNullable(inputDataShape);
  }

  /**
   * @return The value of the {@code outputDataShape} attribute
   */
  @JsonProperty("outputDataShape")
  @Override
  public Optional<DataShape> getOutputDataShape() {
    return Optional.ofNullable(outputDataShape);
  }

  /**
   * @return The value of the {@code propertyDefinitionSteps} attribute
   */
  @JsonProperty("propertyDefinitionSteps")
  @Override
  public List<ActionDescriptor.ActionDescriptorStep> getPropertyDefinitionSteps() {
    return propertyDefinitionSteps;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ExtensionDescriptor#getKind() kind} attribute.
   * A value equality check is used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for kind (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtensionDescriptor withKind(ExtensionAction.Kind value) {
    if (this.kind == value) return this;
    return validate(new ImmutableExtensionDescriptor(
        this,
        value,
        this.entrypoint,
        this.inputDataShape,
        this.outputDataShape,
        this.propertyDefinitionSteps));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ExtensionDescriptor#getEntrypoint() entrypoint} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for entrypoint (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtensionDescriptor withEntrypoint(String value) {
    if (Objects.equals(this.entrypoint, value)) return this;
    return validate(new ImmutableExtensionDescriptor(this, this.kind, value, this.inputDataShape, this.outputDataShape, this.propertyDefinitionSteps));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link ExtensionDescriptor#getInputDataShape() inputDataShape} attribute.
   * @param value The value for inputDataShape
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionDescriptor withInputDataShape(DataShape value) {
    DataShape newValue = Objects.requireNonNull(value, "inputDataShape");
    if (this.inputDataShape == newValue) return this;
    return validate(new ImmutableExtensionDescriptor(this, this.kind, this.entrypoint, newValue, this.outputDataShape, this.propertyDefinitionSteps));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link ExtensionDescriptor#getInputDataShape() inputDataShape} attribute.
   * A shallow reference equality check is used on unboxed optional value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for inputDataShape
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionDescriptor withInputDataShape(Optional<? extends DataShape> optional) {
    DataShape value = optional.orElse(null);
    if (this.inputDataShape == value) return this;
    return validate(new ImmutableExtensionDescriptor(this, this.kind, this.entrypoint, value, this.outputDataShape, this.propertyDefinitionSteps));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link ExtensionDescriptor#getOutputDataShape() outputDataShape} attribute.
   * @param value The value for outputDataShape
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionDescriptor withOutputDataShape(DataShape value) {
    DataShape newValue = Objects.requireNonNull(value, "outputDataShape");
    if (this.outputDataShape == newValue) return this;
    return validate(new ImmutableExtensionDescriptor(this, this.kind, this.entrypoint, this.inputDataShape, newValue, this.propertyDefinitionSteps));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link ExtensionDescriptor#getOutputDataShape() outputDataShape} attribute.
   * A shallow reference equality check is used on unboxed optional value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for outputDataShape
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionDescriptor withOutputDataShape(Optional<? extends DataShape> optional) {
    DataShape value = optional.orElse(null);
    if (this.outputDataShape == value) return this;
    return validate(new ImmutableExtensionDescriptor(this, this.kind, this.entrypoint, this.inputDataShape, value, this.propertyDefinitionSteps));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ExtensionDescriptor#getPropertyDefinitionSteps() propertyDefinitionSteps}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionDescriptor withPropertyDefinitionSteps(ActionDescriptor.ActionDescriptorStep... elements) {
    List<ActionDescriptor.ActionDescriptorStep> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableExtensionDescriptor(this, this.kind, this.entrypoint, this.inputDataShape, this.outputDataShape, newValue));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ExtensionDescriptor#getPropertyDefinitionSteps() propertyDefinitionSteps}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of propertyDefinitionSteps elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionDescriptor withPropertyDefinitionSteps(Iterable<? extends ActionDescriptor.ActionDescriptorStep> elements) {
    if (this.propertyDefinitionSteps == elements) return this;
    List<ActionDescriptor.ActionDescriptorStep> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableExtensionDescriptor(this, this.kind, this.entrypoint, this.inputDataShape, this.outputDataShape, newValue));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableExtensionDescriptor} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableExtensionDescriptor
        && equalTo((ImmutableExtensionDescriptor) another);
  }

  private boolean equalTo(ImmutableExtensionDescriptor another) {
    return Objects.equals(kind, another.kind)
        && Objects.equals(entrypoint, another.entrypoint)
        && Objects.equals(inputDataShape, another.inputDataShape)
        && Objects.equals(outputDataShape, another.outputDataShape)
        && propertyDefinitionSteps.equals(another.propertyDefinitionSteps);
  }

  /**
   * Computes a hash code from attributes: {@code kind}, {@code entrypoint}, {@code inputDataShape}, {@code outputDataShape}, {@code propertyDefinitionSteps}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + Objects.hashCode(kind);
    h += (h << 5) + Objects.hashCode(entrypoint);
    h += (h << 5) + Objects.hashCode(inputDataShape);
    h += (h << 5) + Objects.hashCode(outputDataShape);
    h += (h << 5) + propertyDefinitionSteps.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code ExtensionDescriptor} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder("ExtensionDescriptor{");
    if (kind != null) {
      builder.append("kind=").append(kind);
    }
    if (entrypoint != null) {
      if (builder.length() > 20) builder.append(", ");
      builder.append("entrypoint=").append(entrypoint);
    }
    if (inputDataShape != null) {
      if (builder.length() > 20) builder.append(", ");
      builder.append("inputDataShape=").append(inputDataShape);
    }
    if (outputDataShape != null) {
      if (builder.length() > 20) builder.append(", ");
      builder.append("outputDataShape=").append(outputDataShape);
    }
    if (builder.length() > 20) builder.append(", ");
    builder.append("propertyDefinitionSteps=").append(propertyDefinitionSteps);
    return builder.append("}").toString();
  }

  /**
   * Construct a new immutable {@code ExtensionDescriptor} instance.
   * @param kind The value for the {@code kind} attribute
   * @param entrypoint The value for the {@code entrypoint} attribute
   * @param inputDataShape The value for the {@code inputDataShape} attribute
   * @param outputDataShape The value for the {@code outputDataShape} attribute
   * @param propertyDefinitionSteps The value for the {@code propertyDefinitionSteps} attribute
   * @return An immutable ExtensionDescriptor instance
   */
  public static ExtensionDescriptor of(ExtensionAction.Kind kind, String entrypoint, Optional<DataShape> inputDataShape, Optional<DataShape> outputDataShape, List<ActionDescriptor.ActionDescriptorStep> propertyDefinitionSteps) {
    return of(kind, entrypoint, inputDataShape, outputDataShape, (Iterable<? extends ActionDescriptor.ActionDescriptorStep>) propertyDefinitionSteps);
  }

  /**
   * Construct a new immutable {@code ExtensionDescriptor} instance.
   * @param kind The value for the {@code kind} attribute
   * @param entrypoint The value for the {@code entrypoint} attribute
   * @param inputDataShape The value for the {@code inputDataShape} attribute
   * @param outputDataShape The value for the {@code outputDataShape} attribute
   * @param propertyDefinitionSteps The value for the {@code propertyDefinitionSteps} attribute
   * @return An immutable ExtensionDescriptor instance
   */
  public static ExtensionDescriptor of(ExtensionAction.Kind kind, String entrypoint, Optional<DataShape> inputDataShape, Optional<DataShape> outputDataShape, Iterable<? extends ActionDescriptor.ActionDescriptorStep> propertyDefinitionSteps) {
    return validate(new ImmutableExtensionDescriptor(kind, entrypoint, inputDataShape, outputDataShape, propertyDefinitionSteps));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableExtensionDescriptor validate(ImmutableExtensionDescriptor instance) {
    Set<ConstraintViolation<ImmutableExtensionDescriptor>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link ExtensionDescriptor} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable ExtensionDescriptor instance
   */
  public static ExtensionDescriptor copyOf(ExtensionDescriptor instance) {
    if (instance instanceof ImmutableExtensionDescriptor) {
      return (ImmutableExtensionDescriptor) instance;
    }
    return new ExtensionDescriptor.Builder()
        .createFrom(instance)
        .build();
  }

  private Object readResolve() throws ObjectStreamException {
    return validate(this);
  }

  /**
   * Builds instances of type {@link ExtensionDescriptor ExtensionDescriptor}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static class Builder {
    private ExtensionAction.Kind kind;
    private String entrypoint;
    private DataShape inputDataShape;
    private DataShape outputDataShape;
    private List<ActionDescriptor.ActionDescriptorStep> propertyDefinitionSteps = new ArrayList<ActionDescriptor.ActionDescriptorStep>();

    /**
     * Creates a builder for {@link ExtensionDescriptor ExtensionDescriptor} instances.
     */
    public Builder() {
      if (!(this instanceof ExtensionDescriptor.Builder)) {
        throw new UnsupportedOperationException("Use: new ExtensionDescriptor.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.action.ActionDescriptor} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExtensionDescriptor.Builder createFrom(ActionDescriptor instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ExtensionDescriptor.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.action.ExtensionDescriptor} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExtensionDescriptor.Builder createFrom(ExtensionDescriptor instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ExtensionDescriptor.Builder) this;
    }

    private void from(Object object) {
      if (object instanceof ActionDescriptor) {
        ActionDescriptor instance = (ActionDescriptor) object;
        Optional<DataShape> inputDataShapeOptional = instance.getInputDataShape();
        if (inputDataShapeOptional.isPresent()) {
          inputDataShape(inputDataShapeOptional);
        }
        Optional<DataShape> outputDataShapeOptional = instance.getOutputDataShape();
        if (outputDataShapeOptional.isPresent()) {
          outputDataShape(outputDataShapeOptional);
        }
        addAllPropertyDefinitionSteps(instance.getPropertyDefinitionSteps());
      }
      if (object instanceof ExtensionDescriptor) {
        ExtensionDescriptor instance = (ExtensionDescriptor) object;
        ExtensionAction.Kind kindValue = instance.getKind();
        if (kindValue != null) {
          kind(kindValue);
        }
        String entrypointValue = instance.getEntrypoint();
        if (entrypointValue != null) {
          entrypoint(entrypointValue);
        }
      }
    }

    /**
     * Initializes the value for the {@link ExtensionDescriptor#getKind() kind} attribute.
     * @param kind The value for kind (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("kind")
    public final ExtensionDescriptor.Builder kind(ExtensionAction.Kind kind) {
      this.kind = kind;
      return (ExtensionDescriptor.Builder) this;
    }

    /**
     * Initializes the value for the {@link ExtensionDescriptor#getEntrypoint() entrypoint} attribute.
     * @param entrypoint The value for entrypoint (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("entrypoint")
    public final ExtensionDescriptor.Builder entrypoint(String entrypoint) {
      this.entrypoint = entrypoint;
      return (ExtensionDescriptor.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExtensionDescriptor#getInputDataShape() inputDataShape} to inputDataShape.
     * @param inputDataShape The value for inputDataShape
     * @return {@code this} builder for chained invocation
     */
    public final ExtensionDescriptor.Builder inputDataShape(DataShape inputDataShape) {
      this.inputDataShape = Objects.requireNonNull(inputDataShape, "inputDataShape");
      return (ExtensionDescriptor.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExtensionDescriptor#getInputDataShape() inputDataShape} to inputDataShape.
     * @param inputDataShape The value for inputDataShape
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("inputDataShape")
    public final ExtensionDescriptor.Builder inputDataShape(Optional<? extends DataShape> inputDataShape) {
      this.inputDataShape = inputDataShape.orElse(null);
      return (ExtensionDescriptor.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExtensionDescriptor#getOutputDataShape() outputDataShape} to outputDataShape.
     * @param outputDataShape The value for outputDataShape
     * @return {@code this} builder for chained invocation
     */
    public final ExtensionDescriptor.Builder outputDataShape(DataShape outputDataShape) {
      this.outputDataShape = Objects.requireNonNull(outputDataShape, "outputDataShape");
      return (ExtensionDescriptor.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExtensionDescriptor#getOutputDataShape() outputDataShape} to outputDataShape.
     * @param outputDataShape The value for outputDataShape
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("outputDataShape")
    public final ExtensionDescriptor.Builder outputDataShape(Optional<? extends DataShape> outputDataShape) {
      this.outputDataShape = outputDataShape.orElse(null);
      return (ExtensionDescriptor.Builder) this;
    }

    /**
     * Adds one element to {@link ExtensionDescriptor#getPropertyDefinitionSteps() propertyDefinitionSteps} list.
     * @param element A propertyDefinitionSteps element
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExtensionDescriptor.Builder addPropertyDefinitionStep(ActionDescriptor.ActionDescriptorStep element) {
      this.propertyDefinitionSteps.add(Objects.requireNonNull(element, "propertyDefinitionSteps element"));
      return (ExtensionDescriptor.Builder) this;
    }

    /**
     * Adds elements to {@link ExtensionDescriptor#getPropertyDefinitionSteps() propertyDefinitionSteps} list.
     * @param elements An array of propertyDefinitionSteps elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExtensionDescriptor.Builder addPropertyDefinitionStep(ActionDescriptor.ActionDescriptorStep... elements) {
      for (ActionDescriptor.ActionDescriptorStep element : elements) {
        this.propertyDefinitionSteps.add(Objects.requireNonNull(element, "propertyDefinitionSteps element"));
      }
      return (ExtensionDescriptor.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link ExtensionDescriptor#getPropertyDefinitionSteps() propertyDefinitionSteps} list.
     * @param elements An iterable of propertyDefinitionSteps elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("propertyDefinitionSteps")
    public final ExtensionDescriptor.Builder propertyDefinitionSteps(Iterable<? extends ActionDescriptor.ActionDescriptorStep> elements) {
      this.propertyDefinitionSteps.clear();
      return addAllPropertyDefinitionSteps(elements);
    }

    /**
     * Adds elements to {@link ExtensionDescriptor#getPropertyDefinitionSteps() propertyDefinitionSteps} list.
     * @param elements An iterable of propertyDefinitionSteps elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExtensionDescriptor.Builder addAllPropertyDefinitionSteps(Iterable<? extends ActionDescriptor.ActionDescriptorStep> elements) {
      for (ActionDescriptor.ActionDescriptorStep element : elements) {
        this.propertyDefinitionSteps.add(Objects.requireNonNull(element, "propertyDefinitionSteps element"));
      }
      return (ExtensionDescriptor.Builder) this;
    }

    /**
     * Builds a new {@link ExtensionDescriptor ExtensionDescriptor}.
     * @return An immutable instance of ExtensionDescriptor
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ExtensionDescriptor build() {
      return ImmutableExtensionDescriptor.validate(new ImmutableExtensionDescriptor(
          null,
          kind,
          entrypoint,
          inputDataShape,
          outputDataShape,
          createUnmodifiableList(true, propertyDefinitionSteps)));
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<T>();
    } else {
      list = new ArrayList<T>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<T>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }
}
