/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.model.connection;

import com.fasterxml.jackson.annotation.JsonProperty;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import io.syndesis.model.Dependency;
import io.syndesis.model.Kind;
import io.syndesis.model.WithConfigurationProperties;
import io.syndesis.model.WithConfiguredProperties;
import io.syndesis.model.WithDependencies;
import io.syndesis.model.WithKind;
import io.syndesis.model.WithName;
import io.syndesis.model.action.ActionsSummary;
import io.syndesis.model.action.ConnectorAction;
import java.io.ObjectStreamException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.OptionalInt;
import java.util.Set;
import javax.annotation.Generated;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;

/**
 * Immutable implementation of {@link Connector}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new Connector.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableConnector.of()}.
 */
@SuppressWarnings({"all"})
@SuppressFBWarnings
@Generated({"Immutables.generator", "Connector"})
final class ImmutableConnector implements Connector {
  private final ConnectorGroup connectorGroup;
  private final String connectorGroupId;
  private final String description;
  private final String icon;
  private final Kind kind;
  private final String componentScheme;
  private final String connectorFactory;
  private final ActionsSummary actionsSummary;
  private final Integer uses;
  private final String id;
  private final List<ConnectorAction> actions;
  private final String name;
  private final Map<String, ConfigurationProperty> properties;
  private final Map<String, String> configuredProperties;
  private final List<Dependency> dependencies;

  private ImmutableConnector(
      Optional<ConnectorGroup> connectorGroup,
      Optional<String> connectorGroupId,
      String description,
      String icon,
      Kind kind,
      Optional<String> componentScheme,
      Optional<String> connectorFactory,
      Optional<ActionsSummary> actionsSummary,
      OptionalInt uses,
      Optional<String> id,
      Iterable<? extends ConnectorAction> actions,
      String name,
      Map<String, ? extends ConfigurationProperty> properties,
      Map<String, ? extends String> configuredProperties,
      Iterable<? extends Dependency> dependencies) {
    this.connectorGroup = connectorGroup.orElse(null);
    this.connectorGroupId = connectorGroupId.orElse(null);
    this.description = description;
    this.icon = icon;
    this.kind = Objects.requireNonNull(kind, "kind");
    this.componentScheme = componentScheme.orElse(null);
    this.connectorFactory = connectorFactory.orElse(null);
    this.actionsSummary = actionsSummary.orElse(null);
    this.uses = uses.isPresent() ? uses.getAsInt() : null;
    this.id = id.orElse(null);
    this.actions = createUnmodifiableList(false, createSafeList(actions, true, false));
    this.name = name;
    this.properties = createUnmodifiableMap(true, false, properties);
    this.configuredProperties = createUnmodifiableMap(true, false, configuredProperties);
    this.dependencies = createUnmodifiableList(false, createSafeList(dependencies, true, false));
    initShim.kind(this.kind);
    initShim.configuredProperties(this.configuredProperties);
    initShim.dependencies(this.dependencies);
    this.initShim = null;
  }

  private ImmutableConnector(ImmutableConnector.Builder builder) {
    this.connectorGroup = builder.connectorGroup;
    this.connectorGroupId = builder.connectorGroupId;
    this.description = builder.description;
    this.icon = builder.icon;
    this.componentScheme = builder.componentScheme;
    this.connectorFactory = builder.connectorFactory;
    this.actionsSummary = builder.actionsSummary;
    this.uses = builder.uses;
    this.id = builder.id;
    this.actions = createUnmodifiableList(true, builder.actions);
    this.name = builder.name;
    this.properties = createUnmodifiableMap(false, false, builder.properties);
    if (builder.kind != null) {
      initShim.kind(builder.kind);
    }
    if (builder.configuredPropertiesIsSet()) {
      initShim.configuredProperties(createUnmodifiableMap(false, false, builder.configuredProperties));
    }
    if (builder.dependenciesIsSet()) {
      initShim.dependencies(createUnmodifiableList(true, builder.dependencies));
    }
    this.kind = initShim.getKind();
    this.configuredProperties = initShim.getConfiguredProperties();
    this.dependencies = initShim.getDependencies();
    this.initShim = null;
  }

  private ImmutableConnector(
      ImmutableConnector original,
      ConnectorGroup connectorGroup,
      String connectorGroupId,
      String description,
      String icon,
      Kind kind,
      String componentScheme,
      String connectorFactory,
      ActionsSummary actionsSummary,
      Integer uses,
      String id,
      List<ConnectorAction> actions,
      String name,
      Map<String, ConfigurationProperty> properties,
      Map<String, String> configuredProperties,
      List<Dependency> dependencies) {
    this.connectorGroup = connectorGroup;
    this.connectorGroupId = connectorGroupId;
    this.description = description;
    this.icon = icon;
    this.kind = kind;
    this.componentScheme = componentScheme;
    this.connectorFactory = connectorFactory;
    this.actionsSummary = actionsSummary;
    this.uses = uses;
    this.id = id;
    this.actions = actions;
    this.name = name;
    this.properties = properties;
    this.configuredProperties = configuredProperties;
    this.dependencies = dependencies;
    this.initShim = null;
  }

  private static final int STAGE_INITIALIZING = -1;
  private static final int STAGE_UNINITIALIZED = 0;
  private static final int STAGE_INITIALIZED = 1;
  private transient volatile InitShim initShim = new InitShim();

  private final class InitShim {
    private Kind kind;
    private int kindBuildStage;

    Kind getKind() {
      if (kindBuildStage == STAGE_INITIALIZING) throw new IllegalStateException(formatInitCycleMessage());
      if (kindBuildStage == STAGE_UNINITIALIZED) {
        kindBuildStage = STAGE_INITIALIZING;
        this.kind = Objects.requireNonNull(getKindInitialize(), "kind");
        kindBuildStage = STAGE_INITIALIZED;
      }
      return this.kind;
    }

    void kind(Kind kind) {
      this.kind = kind;
      kindBuildStage = STAGE_INITIALIZED;
    }
    private Map<String, String> configuredProperties;
    private int configuredPropertiesBuildStage;

    Map<String, String> getConfiguredProperties() {
      if (configuredPropertiesBuildStage == STAGE_INITIALIZING) throw new IllegalStateException(formatInitCycleMessage());
      if (configuredPropertiesBuildStage == STAGE_UNINITIALIZED) {
        configuredPropertiesBuildStage = STAGE_INITIALIZING;
        this.configuredProperties = createUnmodifiableMap(true, false, getConfiguredPropertiesInitialize());
        configuredPropertiesBuildStage = STAGE_INITIALIZED;
      }
      return this.configuredProperties;
    }

    void configuredProperties(Map<String, String> configuredProperties) {
      this.configuredProperties = configuredProperties;
      configuredPropertiesBuildStage = STAGE_INITIALIZED;
    }
    private List<Dependency> dependencies;
    private int dependenciesBuildStage;

    List<Dependency> getDependencies() {
      if (dependenciesBuildStage == STAGE_INITIALIZING) throw new IllegalStateException(formatInitCycleMessage());
      if (dependenciesBuildStage == STAGE_UNINITIALIZED) {
        dependenciesBuildStage = STAGE_INITIALIZING;
        this.dependencies = createUnmodifiableList(false, createSafeList(getDependenciesInitialize(), true, false));
        dependenciesBuildStage = STAGE_INITIALIZED;
      }
      return this.dependencies;
    }

    void dependencies(List<Dependency> dependencies) {
      this.dependencies = dependencies;
      dependenciesBuildStage = STAGE_INITIALIZED;
    }

    private String formatInitCycleMessage() {
      ArrayList<String> attributes = new ArrayList<String>();
      if (kindBuildStage == STAGE_INITIALIZING) attributes.add("kind");
      if (configuredPropertiesBuildStage == STAGE_INITIALIZING) attributes.add("configuredProperties");
      if (dependenciesBuildStage == STAGE_INITIALIZING) attributes.add("dependencies");
      return "Cannot build Connector, attribute initializers form cycle" + attributes;
    }
  }

  private Kind getKindInitialize() {
    return Connector.super.getKind();
  }

  private Map<String, String> getConfiguredPropertiesInitialize() {
    return Connector.super.getConfiguredProperties();
  }

  private List<Dependency> getDependenciesInitialize() {
    return Connector.super.getDependencies();
  }

  /**
   * @return The value of the {@code connectorGroup} attribute
   */
  @JsonProperty("connectorGroup")
  @Override
  public Optional<ConnectorGroup> getConnectorGroup() {
    return Optional.ofNullable(connectorGroup);
  }

  /**
   * @return The value of the {@code connectorGroupId} attribute
   */
  @JsonProperty("connectorGroupId")
  @Override
  public Optional<String> getConnectorGroupId() {
    return Optional.ofNullable(connectorGroupId);
  }

  /**
   * @return The value of the {@code description} attribute
   */
  @JsonProperty("description")
  @Override
  public String getDescription() {
    return description;
  }

  /**
   * @return The value of the {@code icon} attribute
   */
  @JsonProperty("icon")
  @Override
  public String getIcon() {
    return icon;
  }

  /**
   * @return The value of the {@code kind} attribute
   */
  @JsonProperty("kind")
  @Override
  public Kind getKind() {
    InitShim shim = this.initShim;
    return shim != null
        ? shim.getKind()
        : this.kind;
  }

  /**
   * @return The value of the {@code componentScheme} attribute
   */
  @JsonProperty("componentScheme")
  @Override
  public Optional<String> getComponentScheme() {
    return Optional.ofNullable(componentScheme);
  }

  /**
   * @return The value of the {@code connectorFactory} attribute
   */
  @JsonProperty("connectorFactory")
  @Override
  public Optional<String> getConnectorFactory() {
    return Optional.ofNullable(connectorFactory);
  }

  /**
   * @return The value of the {@code actionsSummary} attribute
   */
  @JsonProperty("actionsSummary")
  @Override
  public Optional<ActionsSummary> getActionsSummary() {
    return Optional.ofNullable(actionsSummary);
  }

  /**
   * @return The value of the {@code uses} attribute
   */
  @JsonProperty("uses")
  @Override
  public OptionalInt getUses() {
    return uses != null
        ? OptionalInt.of(uses)
        : OptionalInt.empty();
  }

  /**
   * @return The value of the {@code id} attribute
   */
  @JsonProperty("id")
  @Override
  public Optional<String> getId() {
    return Optional.ofNullable(id);
  }

  /**
   * @return The value of the {@code actions} attribute
   */
  @JsonProperty("actions")
  @Override
  public List<ConnectorAction> getActions() {
    return actions;
  }

  /**
   * @return The value of the {@code name} attribute
   */
  @JsonProperty("name")
  @Override
  public String getName() {
    return name;
  }

  /**
   * @return The value of the {@code properties} attribute
   */
  @JsonProperty("properties")
  @Override
  public Map<String, ConfigurationProperty> getProperties() {
    return properties;
  }

  /**
   * @return The value of the {@code configuredProperties} attribute
   */
  @JsonProperty("configuredProperties")
  @Override
  public Map<String, String> getConfiguredProperties() {
    InitShim shim = this.initShim;
    return shim != null
        ? shim.getConfiguredProperties()
        : this.configuredProperties;
  }

  /**
   * @return The value of the {@code dependencies} attribute
   */
  @JsonProperty("dependencies")
  @Override
  public List<Dependency> getDependencies() {
    InitShim shim = this.initShim;
    return shim != null
        ? shim.getDependencies()
        : this.dependencies;
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Connector#getConnectorGroup() connectorGroup} attribute.
   * @param value The value for connectorGroup
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withConnectorGroup(ConnectorGroup value) {
    ConnectorGroup newValue = Objects.requireNonNull(value, "connectorGroup");
    if (this.connectorGroup == newValue) return this;
    return validate(new ImmutableConnector(
        this,
        newValue,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Connector#getConnectorGroup() connectorGroup} attribute.
   * A shallow reference equality check is used on unboxed optional value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for connectorGroup
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withConnectorGroup(Optional<? extends ConnectorGroup> optional) {
    ConnectorGroup value = optional.orElse(null);
    if (this.connectorGroup == value) return this;
    return validate(new ImmutableConnector(
        this,
        value,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Connector#getConnectorGroupId() connectorGroupId} attribute.
   * @param value The value for connectorGroupId
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withConnectorGroupId(String value) {
    String newValue = Objects.requireNonNull(value, "connectorGroupId");
    if (Objects.equals(this.connectorGroupId, newValue)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        newValue,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Connector#getConnectorGroupId() connectorGroupId} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for connectorGroupId
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withConnectorGroupId(Optional<String> optional) {
    String value = optional.orElse(null);
    if (Objects.equals(this.connectorGroupId, value)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        value,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Connector#getDescription() description} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for description (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableConnector withDescription(String value) {
    if (Objects.equals(this.description, value)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        value,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Connector#getIcon() icon} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for icon (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableConnector withIcon(String value) {
    if (Objects.equals(this.icon, value)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        value,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Connector#getKind() kind} attribute.
   * A value equality check is used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for kind
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableConnector withKind(Kind value) {
    if (this.kind == value) return this;
    Kind newValue = Objects.requireNonNull(value, "kind");
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        newValue,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Connector#getComponentScheme() componentScheme} attribute.
   * @param value The value for componentScheme
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withComponentScheme(String value) {
    String newValue = Objects.requireNonNull(value, "componentScheme");
    if (Objects.equals(this.componentScheme, newValue)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        newValue,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Connector#getComponentScheme() componentScheme} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for componentScheme
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withComponentScheme(Optional<String> optional) {
    String value = optional.orElse(null);
    if (Objects.equals(this.componentScheme, value)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        value,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Connector#getConnectorFactory() connectorFactory} attribute.
   * @param value The value for connectorFactory
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withConnectorFactory(String value) {
    String newValue = Objects.requireNonNull(value, "connectorFactory");
    if (Objects.equals(this.connectorFactory, newValue)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        newValue,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Connector#getConnectorFactory() connectorFactory} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for connectorFactory
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withConnectorFactory(Optional<String> optional) {
    String value = optional.orElse(null);
    if (Objects.equals(this.connectorFactory, value)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        value,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Connector#getActionsSummary() actionsSummary} attribute.
   * @param value The value for actionsSummary
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withActionsSummary(ActionsSummary value) {
    ActionsSummary newValue = Objects.requireNonNull(value, "actionsSummary");
    if (this.actionsSummary == newValue) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        newValue,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Connector#getActionsSummary() actionsSummary} attribute.
   * A shallow reference equality check is used on unboxed optional value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for actionsSummary
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withActionsSummary(Optional<? extends ActionsSummary> optional) {
    ActionsSummary value = optional.orElse(null);
    if (this.actionsSummary == value) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        value,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Connector#getUses() uses} attribute.
   * @param value The value for uses
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withUses(int value) {
    Integer newValue = value;
    if (Objects.equals(this.uses, newValue)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        newValue,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Connector#getUses() uses} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for uses
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withUses(OptionalInt optional) {
    Integer value = optional.isPresent() ? optional.getAsInt() : null;
    if (Objects.equals(this.uses, value)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        value,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Connector#getId() id} attribute.
   * @param value The value for id
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withId(String value) {
    String newValue = Objects.requireNonNull(value, "id");
    if (Objects.equals(this.id, newValue)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        newValue,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Connector#getId() id} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for id
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withId(Optional<String> optional) {
    String value = optional.orElse(null);
    if (Objects.equals(this.id, value)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        value,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Connector#getActions() actions}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withActions(ConnectorAction... elements) {
    List<ConnectorAction> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        newValue,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Connector#getActions() actions}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of actions elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withActions(Iterable<? extends ConnectorAction> elements) {
    if (this.actions == elements) return this;
    List<ConnectorAction> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        newValue,
        this.name,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Connector#getName() name} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for name (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableConnector withName(String value) {
    if (Objects.equals(this.name, value)) return this;
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        value,
        this.properties,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by replacing the {@link Connector#getProperties() properties} map with the specified map.
   * Nulls are not permitted as keys or values.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param entries The entries to be added to the properties map
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withProperties(Map<String, ? extends ConfigurationProperty> entries) {
    if (this.properties == entries) return this;
    Map<String, ConfigurationProperty> newValue = createUnmodifiableMap(true, false, entries);
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        newValue,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by replacing the {@link Connector#getConfiguredProperties() configuredProperties} map with the specified map.
   * Nulls are not permitted as keys or values.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param entries The entries to be added to the configuredProperties map
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withConfiguredProperties(Map<String, ? extends String> entries) {
    if (this.configuredProperties == entries) return this;
    Map<String, String> newValue = createUnmodifiableMap(true, false, entries);
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        newValue,
        this.dependencies));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Connector#getDependencies() dependencies}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withDependencies(Dependency... elements) {
    List<Dependency> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        newValue));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Connector#getDependencies() dependencies}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of dependencies elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnector withDependencies(Iterable<? extends Dependency> elements) {
    if (this.dependencies == elements) return this;
    List<Dependency> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableConnector(
        this,
        this.connectorGroup,
        this.connectorGroupId,
        this.description,
        this.icon,
        this.kind,
        this.componentScheme,
        this.connectorFactory,
        this.actionsSummary,
        this.uses,
        this.id,
        this.actions,
        this.name,
        this.properties,
        this.configuredProperties,
        newValue));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableConnector} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableConnector
        && equalTo((ImmutableConnector) another);
  }

  private boolean equalTo(ImmutableConnector another) {
    return Objects.equals(connectorGroup, another.connectorGroup)
        && Objects.equals(connectorGroupId, another.connectorGroupId)
        && Objects.equals(description, another.description)
        && Objects.equals(icon, another.icon)
        && kind.equals(another.kind)
        && Objects.equals(componentScheme, another.componentScheme)
        && Objects.equals(connectorFactory, another.connectorFactory)
        && Objects.equals(actionsSummary, another.actionsSummary)
        && Objects.equals(uses, another.uses)
        && Objects.equals(id, another.id)
        && actions.equals(another.actions)
        && Objects.equals(name, another.name)
        && properties.equals(another.properties)
        && configuredProperties.equals(another.configuredProperties)
        && dependencies.equals(another.dependencies);
  }

  /**
   * Computes a hash code from attributes: {@code connectorGroup}, {@code connectorGroupId}, {@code description}, {@code icon}, {@code kind}, {@code componentScheme}, {@code connectorFactory}, {@code actionsSummary}, {@code uses}, {@code id}, {@code actions}, {@code name}, {@code properties}, {@code configuredProperties}, {@code dependencies}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + Objects.hashCode(connectorGroup);
    h += (h << 5) + Objects.hashCode(connectorGroupId);
    h += (h << 5) + Objects.hashCode(description);
    h += (h << 5) + Objects.hashCode(icon);
    h += (h << 5) + kind.hashCode();
    h += (h << 5) + Objects.hashCode(componentScheme);
    h += (h << 5) + Objects.hashCode(connectorFactory);
    h += (h << 5) + Objects.hashCode(actionsSummary);
    h += (h << 5) + Objects.hashCode(uses);
    h += (h << 5) + Objects.hashCode(id);
    h += (h << 5) + actions.hashCode();
    h += (h << 5) + Objects.hashCode(name);
    h += (h << 5) + properties.hashCode();
    h += (h << 5) + configuredProperties.hashCode();
    h += (h << 5) + dependencies.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code Connector} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder("Connector{");
    if (connectorGroup != null) {
      builder.append("connectorGroup=").append(connectorGroup);
    }
    if (connectorGroupId != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("connectorGroupId=").append(connectorGroupId);
    }
    if (description != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("description=").append(description);
    }
    if (icon != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("icon=").append(icon);
    }
    if (builder.length() > 10) builder.append(", ");
    builder.append("kind=").append(kind);
    if (componentScheme != null) {
      builder.append(", ");
      builder.append("componentScheme=").append(componentScheme);
    }
    if (connectorFactory != null) {
      builder.append(", ");
      builder.append("connectorFactory=").append(connectorFactory);
    }
    if (actionsSummary != null) {
      builder.append(", ");
      builder.append("actionsSummary=").append(actionsSummary);
    }
    if (uses != null) {
      builder.append(", ");
      builder.append("uses=").append(uses);
    }
    if (id != null) {
      builder.append(", ");
      builder.append("id=").append(id);
    }
    builder.append(", ");
    builder.append("actions=").append(actions);
    if (name != null) {
      builder.append(", ");
      builder.append("name=").append(name);
    }
    builder.append(", ");
    builder.append("properties=").append(properties);
    builder.append(", ");
    builder.append("configuredProperties=").append(configuredProperties);
    builder.append(", ");
    builder.append("dependencies=").append(dependencies);
    return builder.append("}").toString();
  }

  /**
   * Construct a new immutable {@code Connector} instance.
   * @param connectorGroup The value for the {@code connectorGroup} attribute
   * @param connectorGroupId The value for the {@code connectorGroupId} attribute
   * @param description The value for the {@code description} attribute
   * @param icon The value for the {@code icon} attribute
   * @param kind The value for the {@code kind} attribute
   * @param componentScheme The value for the {@code componentScheme} attribute
   * @param connectorFactory The value for the {@code connectorFactory} attribute
   * @param actionsSummary The value for the {@code actionsSummary} attribute
   * @param uses The value for the {@code uses} attribute
   * @param id The value for the {@code id} attribute
   * @param actions The value for the {@code actions} attribute
   * @param name The value for the {@code name} attribute
   * @param properties The value for the {@code properties} attribute
   * @param configuredProperties The value for the {@code configuredProperties} attribute
   * @param dependencies The value for the {@code dependencies} attribute
   * @return An immutable Connector instance
   */
  public static Connector of(Optional<ConnectorGroup> connectorGroup, Optional<String> connectorGroupId, String description, String icon, Kind kind, Optional<String> componentScheme, Optional<String> connectorFactory, Optional<ActionsSummary> actionsSummary, OptionalInt uses, Optional<String> id, List<ConnectorAction> actions, String name, Map<String, ConfigurationProperty> properties, Map<String, String> configuredProperties, List<Dependency> dependencies) {
    return of(connectorGroup, connectorGroupId, description, icon, kind, componentScheme, connectorFactory, actionsSummary, uses, id, (Iterable<? extends ConnectorAction>) actions, name, properties, configuredProperties, (Iterable<? extends Dependency>) dependencies);
  }

  /**
   * Construct a new immutable {@code Connector} instance.
   * @param connectorGroup The value for the {@code connectorGroup} attribute
   * @param connectorGroupId The value for the {@code connectorGroupId} attribute
   * @param description The value for the {@code description} attribute
   * @param icon The value for the {@code icon} attribute
   * @param kind The value for the {@code kind} attribute
   * @param componentScheme The value for the {@code componentScheme} attribute
   * @param connectorFactory The value for the {@code connectorFactory} attribute
   * @param actionsSummary The value for the {@code actionsSummary} attribute
   * @param uses The value for the {@code uses} attribute
   * @param id The value for the {@code id} attribute
   * @param actions The value for the {@code actions} attribute
   * @param name The value for the {@code name} attribute
   * @param properties The value for the {@code properties} attribute
   * @param configuredProperties The value for the {@code configuredProperties} attribute
   * @param dependencies The value for the {@code dependencies} attribute
   * @return An immutable Connector instance
   */
  public static Connector of(Optional<ConnectorGroup> connectorGroup, Optional<String> connectorGroupId, String description, String icon, Kind kind, Optional<String> componentScheme, Optional<String> connectorFactory, Optional<ActionsSummary> actionsSummary, OptionalInt uses, Optional<String> id, Iterable<? extends ConnectorAction> actions, String name, Map<String, ? extends ConfigurationProperty> properties, Map<String, ? extends String> configuredProperties, Iterable<? extends Dependency> dependencies) {
    return validate(new ImmutableConnector(connectorGroup, connectorGroupId, description, icon, kind, componentScheme, connectorFactory, actionsSummary, uses, id, actions, name, properties, configuredProperties, dependencies));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableConnector validate(ImmutableConnector instance) {
    Set<ConstraintViolation<ImmutableConnector>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link Connector} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Connector instance
   */
  public static Connector copyOf(Connector instance) {
    if (instance instanceof ImmutableConnector) {
      return (ImmutableConnector) instance;
    }
    return new Connector.Builder()
        .createFrom(instance)
        .build();
  }

  private Object readResolve() throws ObjectStreamException {
    return validate(this);
  }

  /**
   * Builds instances of type {@link Connector Connector}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static class Builder {
    private static final long OPT_BIT_CONFIGURED_PROPERTIES = 0x1L;
    private static final long OPT_BIT_DEPENDENCIES = 0x2L;
    private long optBits;

    private ConnectorGroup connectorGroup;
    private String connectorGroupId;
    private String description;
    private String icon;
    private Kind kind;
    private String componentScheme;
    private String connectorFactory;
    private ActionsSummary actionsSummary;
    private Integer uses;
    private String id;
    private List<ConnectorAction> actions = new ArrayList<ConnectorAction>();
    private String name;
    private Map<String, ConfigurationProperty> properties = new LinkedHashMap<String, ConfigurationProperty>();
    private Map<String, String> configuredProperties = new LinkedHashMap<String, String>();
    private List<Dependency> dependencies = new ArrayList<Dependency>();

    /**
     * Creates a builder for {@link Connector Connector} instances.
     */
    public Builder() {
      if (!(this instanceof Connector.Builder)) {
        throw new UnsupportedOperationException("Use: new Connector.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithKind} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder createFrom(WithKind instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (Connector.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithConfigurationProperties} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder createFrom(WithConfigurationProperties instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (Connector.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.connection.Connector} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder createFrom(Connector instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (Connector.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithDependencies} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder createFrom(WithDependencies instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (Connector.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithConfiguredProperties} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder createFrom(WithConfiguredProperties instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (Connector.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithName} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder createFrom(WithName instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (Connector.Builder) this;
    }

    private void from(Object object) {
      long bits = 0;
      if (object instanceof WithKind) {
        WithKind instance = (WithKind) object;
        if ((bits & 0x1L) == 0) {
          kind(instance.getKind());
          bits |= 0x1L;
        }
      }
      if (object instanceof WithConfigurationProperties) {
        WithConfigurationProperties instance = (WithConfigurationProperties) object;
        putAllProperties(instance.getProperties());
      }
      if (object instanceof Connector) {
        Connector instance = (Connector) object;
        Optional<ConnectorGroup> connectorGroupOptional = instance.getConnectorGroup();
        if (connectorGroupOptional.isPresent()) {
          connectorGroup(connectorGroupOptional);
        }
        Optional<String> connectorFactoryOptional = instance.getConnectorFactory();
        if (connectorFactoryOptional.isPresent()) {
          connectorFactory(connectorFactoryOptional);
        }
        if ((bits & 0x1L) == 0) {
          kind(instance.getKind());
          bits |= 0x1L;
        }
        Optional<String> connectorGroupIdOptional = instance.getConnectorGroupId();
        if (connectorGroupIdOptional.isPresent()) {
          connectorGroupId(connectorGroupIdOptional);
        }
        String iconValue = instance.getIcon();
        if (iconValue != null) {
          icon(iconValue);
        }
        String descriptionValue = instance.getDescription();
        if (descriptionValue != null) {
          description(descriptionValue);
        }
        Optional<String> componentSchemeOptional = instance.getComponentScheme();
        if (componentSchemeOptional.isPresent()) {
          componentScheme(componentSchemeOptional);
        }
        OptionalInt usesOptional = instance.getUses();
        if (usesOptional.isPresent()) {
          uses(usesOptional);
        }
        Optional<String> idOptional = instance.getId();
        if (idOptional.isPresent()) {
          id(idOptional);
        }
        addAllActions(instance.getActions());
        Optional<ActionsSummary> actionsSummaryOptional = instance.getActionsSummary();
        if (actionsSummaryOptional.isPresent()) {
          actionsSummary(actionsSummaryOptional);
        }
      }
      if (object instanceof WithDependencies) {
        WithDependencies instance = (WithDependencies) object;
        addAllDependencies(instance.getDependencies());
      }
      if (object instanceof WithConfiguredProperties) {
        WithConfiguredProperties instance = (WithConfiguredProperties) object;
        putAllConfiguredProperties(instance.getConfiguredProperties());
      }
      if (object instanceof WithName) {
        WithName instance = (WithName) object;
        String nameValue = instance.getName();
        if (nameValue != null) {
          name(nameValue);
        }
      }
    }

    /**
     * Initializes the optional value {@link Connector#getConnectorGroup() connectorGroup} to connectorGroup.
     * @param connectorGroup The value for connectorGroup
     * @return {@code this} builder for chained invocation
     */
    public final Connector.Builder connectorGroup(ConnectorGroup connectorGroup) {
      this.connectorGroup = Objects.requireNonNull(connectorGroup, "connectorGroup");
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getConnectorGroup() connectorGroup} to connectorGroup.
     * @param connectorGroup The value for connectorGroup
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("connectorGroup")
    public final Connector.Builder connectorGroup(Optional<? extends ConnectorGroup> connectorGroup) {
      this.connectorGroup = connectorGroup.orElse(null);
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getConnectorGroupId() connectorGroupId} to connectorGroupId.
     * @param connectorGroupId The value for connectorGroupId
     * @return {@code this} builder for chained invocation
     */
    public final Connector.Builder connectorGroupId(String connectorGroupId) {
      this.connectorGroupId = Objects.requireNonNull(connectorGroupId, "connectorGroupId");
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getConnectorGroupId() connectorGroupId} to connectorGroupId.
     * @param connectorGroupId The value for connectorGroupId
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("connectorGroupId")
    public final Connector.Builder connectorGroupId(Optional<String> connectorGroupId) {
      this.connectorGroupId = connectorGroupId.orElse(null);
      return (Connector.Builder) this;
    }

    /**
     * Initializes the value for the {@link Connector#getDescription() description} attribute.
     * @param description The value for description (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("description")
    public final Connector.Builder description(String description) {
      this.description = description;
      return (Connector.Builder) this;
    }

    /**
     * Initializes the value for the {@link Connector#getIcon() icon} attribute.
     * @param icon The value for icon (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("icon")
    public final Connector.Builder icon(String icon) {
      this.icon = icon;
      return (Connector.Builder) this;
    }

    /**
     * Initializes the value for the {@link Connector#getKind() kind} attribute.
     * <p><em>If not set, this attribute will have a default value as returned by the initializer of {@link Connector#getKind() kind}.</em>
     * @param kind The value for kind 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("kind")
    public final Connector.Builder kind(Kind kind) {
      this.kind = Objects.requireNonNull(kind, "kind");
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getComponentScheme() componentScheme} to componentScheme.
     * @param componentScheme The value for componentScheme
     * @return {@code this} builder for chained invocation
     */
    public final Connector.Builder componentScheme(String componentScheme) {
      this.componentScheme = Objects.requireNonNull(componentScheme, "componentScheme");
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getComponentScheme() componentScheme} to componentScheme.
     * @param componentScheme The value for componentScheme
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("componentScheme")
    public final Connector.Builder componentScheme(Optional<String> componentScheme) {
      this.componentScheme = componentScheme.orElse(null);
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getConnectorFactory() connectorFactory} to connectorFactory.
     * @param connectorFactory The value for connectorFactory
     * @return {@code this} builder for chained invocation
     */
    public final Connector.Builder connectorFactory(String connectorFactory) {
      this.connectorFactory = Objects.requireNonNull(connectorFactory, "connectorFactory");
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getConnectorFactory() connectorFactory} to connectorFactory.
     * @param connectorFactory The value for connectorFactory
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("connectorFactory")
    public final Connector.Builder connectorFactory(Optional<String> connectorFactory) {
      this.connectorFactory = connectorFactory.orElse(null);
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getActionsSummary() actionsSummary} to actionsSummary.
     * @param actionsSummary The value for actionsSummary
     * @return {@code this} builder for chained invocation
     */
    public final Connector.Builder actionsSummary(ActionsSummary actionsSummary) {
      this.actionsSummary = Objects.requireNonNull(actionsSummary, "actionsSummary");
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getActionsSummary() actionsSummary} to actionsSummary.
     * @param actionsSummary The value for actionsSummary
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("actionsSummary")
    public final Connector.Builder actionsSummary(Optional<? extends ActionsSummary> actionsSummary) {
      this.actionsSummary = actionsSummary.orElse(null);
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getUses() uses} to uses.
     * @param uses The value for uses
     * @return {@code this} builder for chained invocation
     */
    public final Connector.Builder uses(int uses) {
      this.uses = uses;
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getUses() uses} to uses.
     * @param uses The value for uses
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("uses")
    public final Connector.Builder uses(OptionalInt uses) {
      this.uses = uses.isPresent() ? uses.getAsInt() : null;
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getId() id} to id.
     * @param id The value for id
     * @return {@code this} builder for chained invocation
     */
    public final Connector.Builder id(String id) {
      this.id = Objects.requireNonNull(id, "id");
      return (Connector.Builder) this;
    }

    /**
     * Initializes the optional value {@link Connector#getId() id} to id.
     * @param id The value for id
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("id")
    public final Connector.Builder id(Optional<String> id) {
      this.id = id.orElse(null);
      return (Connector.Builder) this;
    }

    /**
     * Adds one element to {@link Connector#getActions() actions} list.
     * @param element A actions element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder addAction(ConnectorAction element) {
      this.actions.add(Objects.requireNonNull(element, "actions element"));
      return (Connector.Builder) this;
    }

    /**
     * Adds elements to {@link Connector#getActions() actions} list.
     * @param elements An array of actions elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder addAction(ConnectorAction... elements) {
      for (ConnectorAction element : elements) {
        this.actions.add(Objects.requireNonNull(element, "actions element"));
      }
      return (Connector.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link Connector#getActions() actions} list.
     * @param elements An iterable of actions elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("actions")
    public final Connector.Builder actions(Iterable<? extends ConnectorAction> elements) {
      this.actions.clear();
      return addAllActions(elements);
    }

    /**
     * Adds elements to {@link Connector#getActions() actions} list.
     * @param elements An iterable of actions elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder addAllActions(Iterable<? extends ConnectorAction> elements) {
      for (ConnectorAction element : elements) {
        this.actions.add(Objects.requireNonNull(element, "actions element"));
      }
      return (Connector.Builder) this;
    }

    /**
     * Initializes the value for the {@link Connector#getName() name} attribute.
     * @param name The value for name (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("name")
    public final Connector.Builder name(String name) {
      this.name = name;
      return (Connector.Builder) this;
    }

    /**
     * Put one entry to the {@link Connector#getProperties() properties} map.
     * @param key The key in the properties map
     * @param value The associated value in the properties map
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder putProperty(String key, ConfigurationProperty value) {
      this.properties.put(
          Objects.requireNonNull(key, "properties key"),
          Objects.requireNonNull(value, "properties value"));
      return (Connector.Builder) this;
    }

    /**
     * Put one entry to the {@link Connector#getProperties() properties} map. Nulls are not permitted
     * @param entry The key and value entry
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder putProperty(Map.Entry<String, ? extends ConfigurationProperty> entry) {
      String k = entry.getKey();
      ConfigurationProperty v = entry.getValue();
      this.properties.put(
          Objects.requireNonNull(k, "properties key"),
          Objects.requireNonNull(v, "properties value"));
      return (Connector.Builder) this;
    }

    /**
     * Sets or replaces all mappings from the specified map as entries for the {@link Connector#getProperties() properties} map. Nulls are not permitted
     * @param properties The entries that will be added to the properties map
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("properties")
    public final Connector.Builder properties(Map<String, ? extends ConfigurationProperty> properties) {
      this.properties.clear();
      return putAllProperties(properties);
    }

    /**
     * Put all mappings from the specified map as entries to {@link Connector#getProperties() properties} map. Nulls are not permitted
     * @param properties The entries that will be added to the properties map
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder putAllProperties(Map<String, ? extends ConfigurationProperty> properties) {
      for (Map.Entry<String, ? extends ConfigurationProperty> entry : properties.entrySet()) {
        String k = entry.getKey();
        ConfigurationProperty v = entry.getValue();
        this.properties.put(
            Objects.requireNonNull(k, "properties key"),
            Objects.requireNonNull(v, "properties value"));
      }
      return (Connector.Builder) this;
    }

    /**
     * Put one entry to the {@link Connector#getConfiguredProperties() configuredProperties} map.
     * @param key The key in the configuredProperties map
     * @param value The associated value in the configuredProperties map
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder putConfiguredProperty(String key, String value) {
      this.configuredProperties.put(
          Objects.requireNonNull(key, "configuredProperties key"),
          Objects.requireNonNull(value, "configuredProperties value"));
      optBits |= OPT_BIT_CONFIGURED_PROPERTIES;
      return (Connector.Builder) this;
    }

    /**
     * Put one entry to the {@link Connector#getConfiguredProperties() configuredProperties} map. Nulls are not permitted
     * @param entry The key and value entry
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder putConfiguredProperty(Map.Entry<String, ? extends String> entry) {
      String k = entry.getKey();
      String v = entry.getValue();
      this.configuredProperties.put(
          Objects.requireNonNull(k, "configuredProperties key"),
          Objects.requireNonNull(v, "configuredProperties value"));
      optBits |= OPT_BIT_CONFIGURED_PROPERTIES;
      return (Connector.Builder) this;
    }

    /**
     * Sets or replaces all mappings from the specified map as entries for the {@link Connector#getConfiguredProperties() configuredProperties} map. Nulls are not permitted
     * @param configuredProperties The entries that will be added to the configuredProperties map
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("configuredProperties")
    public final Connector.Builder configuredProperties(Map<String, ? extends String> configuredProperties) {
      this.configuredProperties.clear();
      optBits |= OPT_BIT_CONFIGURED_PROPERTIES;
      return putAllConfiguredProperties(configuredProperties);
    }

    /**
     * Put all mappings from the specified map as entries to {@link Connector#getConfiguredProperties() configuredProperties} map. Nulls are not permitted
     * @param configuredProperties The entries that will be added to the configuredProperties map
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder putAllConfiguredProperties(Map<String, ? extends String> configuredProperties) {
      for (Map.Entry<String, ? extends String> entry : configuredProperties.entrySet()) {
        String k = entry.getKey();
        String v = entry.getValue();
        this.configuredProperties.put(
            Objects.requireNonNull(k, "configuredProperties key"),
            Objects.requireNonNull(v, "configuredProperties value"));
      }
      optBits |= OPT_BIT_CONFIGURED_PROPERTIES;
      return (Connector.Builder) this;
    }

    /**
     * Adds one element to {@link Connector#getDependencies() dependencies} list.
     * @param element A dependencies element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder addDependency(Dependency element) {
      this.dependencies.add(Objects.requireNonNull(element, "dependencies element"));
      optBits |= OPT_BIT_DEPENDENCIES;
      return (Connector.Builder) this;
    }

    /**
     * Adds elements to {@link Connector#getDependencies() dependencies} list.
     * @param elements An array of dependencies elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder addDependency(Dependency... elements) {
      for (Dependency element : elements) {
        this.dependencies.add(Objects.requireNonNull(element, "dependencies element"));
      }
      optBits |= OPT_BIT_DEPENDENCIES;
      return (Connector.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link Connector#getDependencies() dependencies} list.
     * @param elements An iterable of dependencies elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("dependencies")
    public final Connector.Builder dependencies(Iterable<? extends Dependency> elements) {
      this.dependencies.clear();
      return addAllDependencies(elements);
    }

    /**
     * Adds elements to {@link Connector#getDependencies() dependencies} list.
     * @param elements An iterable of dependencies elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Connector.Builder addAllDependencies(Iterable<? extends Dependency> elements) {
      for (Dependency element : elements) {
        this.dependencies.add(Objects.requireNonNull(element, "dependencies element"));
      }
      optBits |= OPT_BIT_DEPENDENCIES;
      return (Connector.Builder) this;
    }

    /**
     * Builds a new {@link Connector Connector}.
     * @return An immutable instance of Connector
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public Connector build() {
      return ImmutableConnector.validate(new ImmutableConnector(this));
    }

    private boolean configuredPropertiesIsSet() {
      return (optBits & OPT_BIT_CONFIGURED_PROPERTIES) != 0;
    }

    private boolean dependenciesIsSet() {
      return (optBits & OPT_BIT_DEPENDENCIES) != 0;
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<T>();
    } else {
      list = new ArrayList<T>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<T>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }

  private static <K, V> Map<K, V> createUnmodifiableMap(boolean checkNulls, boolean skipNulls, Map<? extends K, ? extends V> map) {
    switch (map.size()) {
    case 0: return Collections.emptyMap();
    case 1: {
      Map.Entry<? extends K, ? extends V> e = map.entrySet().iterator().next();
      K k = e.getKey();
      V v = e.getValue();
      if (checkNulls) {
        Objects.requireNonNull(k, "key");
        Objects.requireNonNull(v, "value");
      }
      if (skipNulls && (k == null || v == null)) {
        return Collections.emptyMap();
      }
      return Collections.singletonMap(k, v);
    }
    default: {
      Map<K, V> linkedMap = new LinkedHashMap<K, V>(map.size());
      if (skipNulls || checkNulls) {
        for (Map.Entry<? extends K, ? extends V> e : map.entrySet()) {
          K k = e.getKey();
          V v = e.getValue();
          if (skipNulls) {
            if (k == null || v == null) continue;
          } else if (checkNulls) {
            Objects.requireNonNull(k, "key");
            Objects.requireNonNull(v, "value");
          }
          linkedMap.put(k, v);
        }
      } else {
        linkedMap.putAll(map);
      }
      return Collections.unmodifiableMap(linkedMap);
    }
    }
  }
}
