/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.model.connection;

import com.fasterxml.jackson.annotation.JsonProperty;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import io.syndesis.model.Violation;
import io.syndesis.model.WithConfigurationProperties;
import io.syndesis.model.WithName;
import io.syndesis.model.action.ActionsSummary;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import javax.annotation.Generated;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;

/**
 * Immutable implementation of {@link ConnectorSummary}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new ConnectorSummary.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableConnectorSummary.of()}.
 */
@SuppressWarnings({"all"})
@SuppressFBWarnings
@Generated({"Immutables.generator", "ConnectorSummary"})
final class ImmutableConnectorSummary implements ConnectorSummary {
  private final ActionsSummary actionsSummary;
  private final String description;
  private final List<Violation> errors;
  private final String icon;
  private final List<Violation> warnings;
  private final String name;
  private final Map<String, ConfigurationProperty> properties;

  private ImmutableConnectorSummary(
      ActionsSummary actionsSummary,
      String description,
      Iterable<? extends Violation> errors,
      String icon,
      Iterable<? extends Violation> warnings,
      String name,
      Map<String, ? extends ConfigurationProperty> properties) {
    this.actionsSummary = actionsSummary;
    this.description = description;
    this.errors = createUnmodifiableList(false, createSafeList(errors, true, false));
    this.icon = icon;
    this.warnings = createUnmodifiableList(false, createSafeList(warnings, true, false));
    this.name = name;
    this.properties = createUnmodifiableMap(true, false, properties);
  }

  private ImmutableConnectorSummary(
      ImmutableConnectorSummary original,
      ActionsSummary actionsSummary,
      String description,
      List<Violation> errors,
      String icon,
      List<Violation> warnings,
      String name,
      Map<String, ConfigurationProperty> properties) {
    this.actionsSummary = actionsSummary;
    this.description = description;
    this.errors = errors;
    this.icon = icon;
    this.warnings = warnings;
    this.name = name;
    this.properties = properties;
  }

  /**
   * @return The value of the {@code actionsSummary} attribute
   */
  @JsonProperty("actionsSummary")
  @Override
  public ActionsSummary getActionsSummary() {
    return actionsSummary;
  }

  /**
   * @return The value of the {@code description} attribute
   */
  @JsonProperty("description")
  @Override
  public String getDescription() {
    return description;
  }

  /**
   * @return The value of the {@code errors} attribute
   */
  @JsonProperty("errors")
  @Override
  public List<Violation> getErrors() {
    return errors;
  }

  /**
   * @return The value of the {@code icon} attribute
   */
  @JsonProperty("icon")
  @Override
  public String getIcon() {
    return icon;
  }

  /**
   * @return The value of the {@code warnings} attribute
   */
  @JsonProperty("warnings")
  @Override
  public List<Violation> getWarnings() {
    return warnings;
  }

  /**
   * @return The value of the {@code name} attribute
   */
  @JsonProperty("name")
  @Override
  public String getName() {
    return name;
  }

  /**
   * @return The value of the {@code properties} attribute
   */
  @JsonProperty("properties")
  @Override
  public Map<String, ConfigurationProperty> getProperties() {
    return properties;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ConnectorSummary#getActionsSummary() actionsSummary} attribute.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for actionsSummary (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableConnectorSummary withActionsSummary(ActionsSummary value) {
    if (this.actionsSummary == value) return this;
    return validate(new ImmutableConnectorSummary(
        this,
        value,
        this.description,
        this.errors,
        this.icon,
        this.warnings,
        this.name,
        this.properties));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ConnectorSummary#getDescription() description} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for description (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableConnectorSummary withDescription(String value) {
    if (Objects.equals(this.description, value)) return this;
    return validate(new ImmutableConnectorSummary(
        this,
        this.actionsSummary,
        value,
        this.errors,
        this.icon,
        this.warnings,
        this.name,
        this.properties));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ConnectorSummary#getErrors() errors}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnectorSummary withErrors(Violation... elements) {
    List<Violation> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableConnectorSummary(
        this,
        this.actionsSummary,
        this.description,
        newValue,
        this.icon,
        this.warnings,
        this.name,
        this.properties));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ConnectorSummary#getErrors() errors}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of errors elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnectorSummary withErrors(Iterable<? extends Violation> elements) {
    if (this.errors == elements) return this;
    List<Violation> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableConnectorSummary(
        this,
        this.actionsSummary,
        this.description,
        newValue,
        this.icon,
        this.warnings,
        this.name,
        this.properties));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ConnectorSummary#getIcon() icon} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for icon (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableConnectorSummary withIcon(String value) {
    if (Objects.equals(this.icon, value)) return this;
    return validate(new ImmutableConnectorSummary(
        this,
        this.actionsSummary,
        this.description,
        this.errors,
        value,
        this.warnings,
        this.name,
        this.properties));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ConnectorSummary#getWarnings() warnings}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnectorSummary withWarnings(Violation... elements) {
    List<Violation> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableConnectorSummary(
        this,
        this.actionsSummary,
        this.description,
        this.errors,
        this.icon,
        newValue,
        this.name,
        this.properties));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ConnectorSummary#getWarnings() warnings}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of warnings elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnectorSummary withWarnings(Iterable<? extends Violation> elements) {
    if (this.warnings == elements) return this;
    List<Violation> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableConnectorSummary(
        this,
        this.actionsSummary,
        this.description,
        this.errors,
        this.icon,
        newValue,
        this.name,
        this.properties));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ConnectorSummary#getName() name} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for name (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableConnectorSummary withName(String value) {
    if (Objects.equals(this.name, value)) return this;
    return validate(new ImmutableConnectorSummary(
        this,
        this.actionsSummary,
        this.description,
        this.errors,
        this.icon,
        this.warnings,
        value,
        this.properties));
  }

  /**
   * Copy the current immutable object by replacing the {@link ConnectorSummary#getProperties() properties} map with the specified map.
   * Nulls are not permitted as keys or values.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param entries The entries to be added to the properties map
   * @return A modified copy of {@code this} object
   */
  public final ImmutableConnectorSummary withProperties(Map<String, ? extends ConfigurationProperty> entries) {
    if (this.properties == entries) return this;
    Map<String, ConfigurationProperty> newValue = createUnmodifiableMap(true, false, entries);
    return validate(new ImmutableConnectorSummary(
        this,
        this.actionsSummary,
        this.description,
        this.errors,
        this.icon,
        this.warnings,
        this.name,
        newValue));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableConnectorSummary} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableConnectorSummary
        && equalTo((ImmutableConnectorSummary) another);
  }

  private boolean equalTo(ImmutableConnectorSummary another) {
    return Objects.equals(actionsSummary, another.actionsSummary)
        && Objects.equals(description, another.description)
        && errors.equals(another.errors)
        && Objects.equals(icon, another.icon)
        && warnings.equals(another.warnings)
        && Objects.equals(name, another.name)
        && properties.equals(another.properties);
  }

  /**
   * Computes a hash code from attributes: {@code actionsSummary}, {@code description}, {@code errors}, {@code icon}, {@code warnings}, {@code name}, {@code properties}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + Objects.hashCode(actionsSummary);
    h += (h << 5) + Objects.hashCode(description);
    h += (h << 5) + errors.hashCode();
    h += (h << 5) + Objects.hashCode(icon);
    h += (h << 5) + warnings.hashCode();
    h += (h << 5) + Objects.hashCode(name);
    h += (h << 5) + properties.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code ConnectorSummary} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "ConnectorSummary{"
        + "actionsSummary=" + actionsSummary
        + ", description=" + description
        + ", errors=" + errors
        + ", icon=" + icon
        + ", warnings=" + warnings
        + ", name=" + name
        + ", properties=" + properties
        + "}";
  }

  /**
   * Construct a new immutable {@code ConnectorSummary} instance.
   * @param actionsSummary The value for the {@code actionsSummary} attribute
   * @param description The value for the {@code description} attribute
   * @param errors The value for the {@code errors} attribute
   * @param icon The value for the {@code icon} attribute
   * @param warnings The value for the {@code warnings} attribute
   * @param name The value for the {@code name} attribute
   * @param properties The value for the {@code properties} attribute
   * @return An immutable ConnectorSummary instance
   */
  public static ConnectorSummary of(ActionsSummary actionsSummary, String description, List<Violation> errors, String icon, List<Violation> warnings, String name, Map<String, ConfigurationProperty> properties) {
    return of(actionsSummary, description, (Iterable<? extends Violation>) errors, icon, (Iterable<? extends Violation>) warnings, name, properties);
  }

  /**
   * Construct a new immutable {@code ConnectorSummary} instance.
   * @param actionsSummary The value for the {@code actionsSummary} attribute
   * @param description The value for the {@code description} attribute
   * @param errors The value for the {@code errors} attribute
   * @param icon The value for the {@code icon} attribute
   * @param warnings The value for the {@code warnings} attribute
   * @param name The value for the {@code name} attribute
   * @param properties The value for the {@code properties} attribute
   * @return An immutable ConnectorSummary instance
   */
  public static ConnectorSummary of(ActionsSummary actionsSummary, String description, Iterable<? extends Violation> errors, String icon, Iterable<? extends Violation> warnings, String name, Map<String, ? extends ConfigurationProperty> properties) {
    return validate(new ImmutableConnectorSummary(actionsSummary, description, errors, icon, warnings, name, properties));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableConnectorSummary validate(ImmutableConnectorSummary instance) {
    Set<ConstraintViolation<ImmutableConnectorSummary>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link ConnectorSummary} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable ConnectorSummary instance
   */
  public static ConnectorSummary copyOf(ConnectorSummary instance) {
    if (instance instanceof ImmutableConnectorSummary) {
      return (ImmutableConnectorSummary) instance;
    }
    return new ConnectorSummary.Builder()
        .createFrom(instance)
        .build();
  }

  /**
   * Builds instances of type {@link ConnectorSummary ConnectorSummary}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static class Builder {
    private ActionsSummary actionsSummary;
    private String description;
    private List<Violation> errors = new ArrayList<Violation>();
    private String icon;
    private List<Violation> warnings = new ArrayList<Violation>();
    private String name;
    private Map<String, ConfigurationProperty> properties = new LinkedHashMap<String, ConfigurationProperty>();

    /**
     * Creates a builder for {@link ConnectorSummary ConnectorSummary} instances.
     */
    public Builder() {
      if (!(this instanceof ConnectorSummary.Builder)) {
        throw new UnsupportedOperationException("Use: new ConnectorSummary.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithConfigurationProperties} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder createFrom(WithConfigurationProperties instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.connection.ConnectorSummary} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder createFrom(ConnectorSummary instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithName} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder createFrom(WithName instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ConnectorSummary.Builder) this;
    }

    private void from(Object object) {
      if (object instanceof WithConfigurationProperties) {
        WithConfigurationProperties instance = (WithConfigurationProperties) object;
        putAllProperties(instance.getProperties());
      }
      if (object instanceof ConnectorSummary) {
        ConnectorSummary instance = (ConnectorSummary) object;
        String iconValue = instance.getIcon();
        if (iconValue != null) {
          icon(iconValue);
        }
        String descriptionValue = instance.getDescription();
        if (descriptionValue != null) {
          description(descriptionValue);
        }
        ActionsSummary actionsSummaryValue = instance.getActionsSummary();
        if (actionsSummaryValue != null) {
          actionsSummary(actionsSummaryValue);
        }
        addAllErrors(instance.getErrors());
        addAllWarnings(instance.getWarnings());
      }
      if (object instanceof WithName) {
        WithName instance = (WithName) object;
        String nameValue = instance.getName();
        if (nameValue != null) {
          name(nameValue);
        }
      }
    }

    /**
     * Initializes the value for the {@link ConnectorSummary#getActionsSummary() actionsSummary} attribute.
     * @param actionsSummary The value for actionsSummary (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("actionsSummary")
    public final ConnectorSummary.Builder actionsSummary(ActionsSummary actionsSummary) {
      this.actionsSummary = actionsSummary;
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Initializes the value for the {@link ConnectorSummary#getDescription() description} attribute.
     * @param description The value for description (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("description")
    public final ConnectorSummary.Builder description(String description) {
      this.description = description;
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Adds one element to {@link ConnectorSummary#getErrors() errors} list.
     * @param element A errors element
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder addError(Violation element) {
      this.errors.add(Objects.requireNonNull(element, "errors element"));
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Adds elements to {@link ConnectorSummary#getErrors() errors} list.
     * @param elements An array of errors elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder addError(Violation... elements) {
      for (Violation element : elements) {
        this.errors.add(Objects.requireNonNull(element, "errors element"));
      }
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link ConnectorSummary#getErrors() errors} list.
     * @param elements An iterable of errors elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("errors")
    public final ConnectorSummary.Builder errors(Iterable<? extends Violation> elements) {
      this.errors.clear();
      return addAllErrors(elements);
    }

    /**
     * Adds elements to {@link ConnectorSummary#getErrors() errors} list.
     * @param elements An iterable of errors elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder addAllErrors(Iterable<? extends Violation> elements) {
      for (Violation element : elements) {
        this.errors.add(Objects.requireNonNull(element, "errors element"));
      }
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Initializes the value for the {@link ConnectorSummary#getIcon() icon} attribute.
     * @param icon The value for icon (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("icon")
    public final ConnectorSummary.Builder icon(String icon) {
      this.icon = icon;
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Adds one element to {@link ConnectorSummary#getWarnings() warnings} list.
     * @param element A warnings element
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder addWarning(Violation element) {
      this.warnings.add(Objects.requireNonNull(element, "warnings element"));
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Adds elements to {@link ConnectorSummary#getWarnings() warnings} list.
     * @param elements An array of warnings elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder addWarning(Violation... elements) {
      for (Violation element : elements) {
        this.warnings.add(Objects.requireNonNull(element, "warnings element"));
      }
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link ConnectorSummary#getWarnings() warnings} list.
     * @param elements An iterable of warnings elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("warnings")
    public final ConnectorSummary.Builder warnings(Iterable<? extends Violation> elements) {
      this.warnings.clear();
      return addAllWarnings(elements);
    }

    /**
     * Adds elements to {@link ConnectorSummary#getWarnings() warnings} list.
     * @param elements An iterable of warnings elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder addAllWarnings(Iterable<? extends Violation> elements) {
      for (Violation element : elements) {
        this.warnings.add(Objects.requireNonNull(element, "warnings element"));
      }
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Initializes the value for the {@link ConnectorSummary#getName() name} attribute.
     * @param name The value for name (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("name")
    public final ConnectorSummary.Builder name(String name) {
      this.name = name;
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Put one entry to the {@link ConnectorSummary#getProperties() properties} map.
     * @param key The key in the properties map
     * @param value The associated value in the properties map
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder putProperty(String key, ConfigurationProperty value) {
      this.properties.put(
          Objects.requireNonNull(key, "properties key"),
          Objects.requireNonNull(value, "properties value"));
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Put one entry to the {@link ConnectorSummary#getProperties() properties} map. Nulls are not permitted
     * @param entry The key and value entry
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder putProperty(Map.Entry<String, ? extends ConfigurationProperty> entry) {
      String k = entry.getKey();
      ConfigurationProperty v = entry.getValue();
      this.properties.put(
          Objects.requireNonNull(k, "properties key"),
          Objects.requireNonNull(v, "properties value"));
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Sets or replaces all mappings from the specified map as entries for the {@link ConnectorSummary#getProperties() properties} map. Nulls are not permitted
     * @param properties The entries that will be added to the properties map
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("properties")
    public final ConnectorSummary.Builder properties(Map<String, ? extends ConfigurationProperty> properties) {
      this.properties.clear();
      return putAllProperties(properties);
    }

    /**
     * Put all mappings from the specified map as entries to {@link ConnectorSummary#getProperties() properties} map. Nulls are not permitted
     * @param properties The entries that will be added to the properties map
     * @return {@code this} builder for use in a chained invocation
     */
    public final ConnectorSummary.Builder putAllProperties(Map<String, ? extends ConfigurationProperty> properties) {
      for (Map.Entry<String, ? extends ConfigurationProperty> entry : properties.entrySet()) {
        String k = entry.getKey();
        ConfigurationProperty v = entry.getValue();
        this.properties.put(
            Objects.requireNonNull(k, "properties key"),
            Objects.requireNonNull(v, "properties value"));
      }
      return (ConnectorSummary.Builder) this;
    }

    /**
     * Builds a new {@link ConnectorSummary ConnectorSummary}.
     * @return An immutable instance of ConnectorSummary
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ConnectorSummary build() {
      return ImmutableConnectorSummary.validate(new ImmutableConnectorSummary(
          null,
          actionsSummary,
          description,
          createUnmodifiableList(true, errors),
          icon,
          createUnmodifiableList(true, warnings),
          name,
          createUnmodifiableMap(false, false, properties)));
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<T>();
    } else {
      list = new ArrayList<T>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<T>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }

  private static <K, V> Map<K, V> createUnmodifiableMap(boolean checkNulls, boolean skipNulls, Map<? extends K, ? extends V> map) {
    switch (map.size()) {
    case 0: return Collections.emptyMap();
    case 1: {
      Map.Entry<? extends K, ? extends V> e = map.entrySet().iterator().next();
      K k = e.getKey();
      V v = e.getValue();
      if (checkNulls) {
        Objects.requireNonNull(k, "key");
        Objects.requireNonNull(v, "value");
      }
      if (skipNulls && (k == null || v == null)) {
        return Collections.emptyMap();
      }
      return Collections.singletonMap(k, v);
    }
    default: {
      Map<K, V> linkedMap = new LinkedHashMap<K, V>(map.size());
      if (skipNulls || checkNulls) {
        for (Map.Entry<? extends K, ? extends V> e : map.entrySet()) {
          K k = e.getKey();
          V v = e.getValue();
          if (skipNulls) {
            if (k == null || v == null) continue;
          } else if (checkNulls) {
            Objects.requireNonNull(k, "key");
            Objects.requireNonNull(v, "value");
          }
          linkedMap.put(k, v);
        }
      } else {
        linkedMap.putAll(map);
      }
      return Collections.unmodifiableMap(linkedMap);
    }
    }
  }
}
