/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.model.extension;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonPropertyOrder;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import io.syndesis.model.Dependency;
import io.syndesis.model.WithConfigurationProperties;
import io.syndesis.model.WithDependencies;
import io.syndesis.model.WithName;
import io.syndesis.model.WithTags;
import io.syndesis.model.action.ExtensionAction;
import io.syndesis.model.connection.ConfigurationProperty;
import java.io.ObjectStreamException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.NavigableSet;
import java.util.Objects;
import java.util.Optional;
import java.util.OptionalInt;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import javax.annotation.Generated;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;

/**
 * Immutable implementation of {@link Extension}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new Extension.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableExtension.of()}.
 */
@SuppressWarnings({"all"})
@SuppressFBWarnings
@Generated({"Immutables.generator", "Extension"})
final class ImmutableExtension implements Extension {
  private final String version;
  private final String extensionId;
  private final Extension.Status status;
  private final String icon;
  private final String description;
  private final Integer uses;
  private final String userId;
  private final Date lastUpdated;
  private final Date createdDate;
  private final String id;
  private final List<ExtensionAction> actions;
  private final String name;
  private final SortedSet<String> tags;
  private final Map<String, ConfigurationProperty> properties;
  private final List<Dependency> dependencies;

  private ImmutableExtension(
      String version,
      String extensionId,
      Optional<Extension.Status> status,
      String icon,
      String description,
      OptionalInt uses,
      Optional<String> userId,
      Optional<Date> lastUpdated,
      Optional<Date> createdDate,
      Optional<String> id,
      Iterable<? extends ExtensionAction> actions,
      String name,
      Iterable<String> tags,
      Map<String, ? extends ConfigurationProperty> properties,
      Iterable<? extends Dependency> dependencies) {
    this.version = version;
    this.extensionId = extensionId;
    this.status = status.orElse(null);
    this.icon = icon;
    this.description = description;
    this.uses = uses.isPresent() ? uses.getAsInt() : null;
    this.userId = userId.orElse(null);
    this.lastUpdated = lastUpdated.orElse(null);
    this.createdDate = createdDate.orElse(null);
    this.id = id.orElse(null);
    this.actions = createUnmodifiableList(false, createSafeList(actions, true, false));
    this.name = name;
    this.tags = createUnmodifiableSortedSet(false, createSafeList(tags, false, true));
    this.properties = createUnmodifiableMap(true, false, properties);
    this.dependencies = createUnmodifiableList(false, createSafeList(dependencies, true, false));
  }

  private ImmutableExtension(ImmutableExtension.Builder builder) {
    this.version = builder.version;
    this.extensionId = builder.extensionId;
    this.status = builder.status;
    this.icon = builder.icon;
    this.description = builder.description;
    this.uses = builder.uses;
    this.userId = builder.userId;
    this.lastUpdated = builder.lastUpdated;
    this.createdDate = builder.createdDate;
    this.id = builder.id;
    this.actions = createUnmodifiableList(true, builder.actions);
    this.name = builder.name;
    this.tags = createUnmodifiableSortedSet(false, createSafeList(builder.tags, false, false));
    this.properties = createUnmodifiableMap(false, false, builder.properties);
    this.dependencies = builder.dependenciesIsSet()
        ? createUnmodifiableList(true, builder.dependencies)
        : createUnmodifiableList(false, createSafeList(Extension.super.getDependencies(), true, false));
  }

  private ImmutableExtension(
      ImmutableExtension original,
      String version,
      String extensionId,
      Extension.Status status,
      String icon,
      String description,
      Integer uses,
      String userId,
      Date lastUpdated,
      Date createdDate,
      String id,
      List<ExtensionAction> actions,
      String name,
      SortedSet<String> tags,
      Map<String, ConfigurationProperty> properties,
      List<Dependency> dependencies) {
    this.version = version;
    this.extensionId = extensionId;
    this.status = status;
    this.icon = icon;
    this.description = description;
    this.uses = uses;
    this.userId = userId;
    this.lastUpdated = lastUpdated;
    this.createdDate = createdDate;
    this.id = id;
    this.actions = actions;
    this.name = name;
    this.tags = tags;
    this.properties = properties;
    this.dependencies = dependencies;
  }

  /**
   * @return The value of the {@code version} attribute
   */
  @JsonProperty("version")
  @Override
  public String getVersion() {
    return version;
  }

  /**
   * @return The value of the {@code extensionId} attribute
   */
  @JsonProperty("extensionId")
  @Override
  public String getExtensionId() {
    return extensionId;
  }

  /**
   * @return The value of the {@code status} attribute
   */
  @JsonProperty("status")
  @Override
  public Optional<Extension.Status> getStatus() {
    return Optional.ofNullable(status);
  }

  /**
   * @return The value of the {@code icon} attribute
   */
  @JsonProperty("icon")
  @Override
  public String getIcon() {
    return icon;
  }

  /**
   * @return The value of the {@code description} attribute
   */
  @JsonProperty("description")
  @Override
  public String getDescription() {
    return description;
  }

  /**
   * @return The value of the {@code uses} attribute
   */
  @JsonProperty("uses")
  @Override
  public OptionalInt getUses() {
    return uses != null
        ? OptionalInt.of(uses)
        : OptionalInt.empty();
  }

  /**
   * @return The value of the {@code userId} attribute
   */
  @JsonProperty("userId")
  @Override
  public Optional<String> getUserId() {
    return Optional.ofNullable(userId);
  }

  /**
   * @return The value of the {@code lastUpdated} attribute
   */
  @JsonProperty("lastUpdated")
  @Override
  public Optional<Date> getLastUpdated() {
    return Optional.ofNullable(lastUpdated);
  }

  /**
   * @return The value of the {@code createdDate} attribute
   */
  @JsonProperty("createdDate")
  @Override
  public Optional<Date> getCreatedDate() {
    return Optional.ofNullable(createdDate);
  }

  /**
   * @return The value of the {@code id} attribute
   */
  @JsonProperty("id")
  @Override
  public Optional<String> getId() {
    return Optional.ofNullable(id);
  }

  /**
   * @return The value of the {@code actions} attribute
   */
  @JsonProperty("actions")
  @Override
  public List<ExtensionAction> getActions() {
    return actions;
  }

  /**
   * @return The value of the {@code name} attribute
   */
  @JsonProperty("name")
  @Override
  public String getName() {
    return name;
  }

  /**
   * @return The value of the {@code tags} attribute
   */
  @JsonProperty("tags")
  @Override
  public SortedSet<String> getTags() {
    return tags;
  }

  /**
   * @return The value of the {@code properties} attribute
   */
  @JsonProperty("properties")
  @Override
  public Map<String, ConfigurationProperty> getProperties() {
    return properties;
  }

  /**
   * @return The value of the {@code dependencies} attribute
   */
  @JsonProperty("dependencies")
  @Override
  public List<Dependency> getDependencies() {
    return dependencies;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Extension#getVersion() version} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for version (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtension withVersion(String value) {
    if (Objects.equals(this.version, value)) return this;
    return validate(new ImmutableExtension(
        this,
        value,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Extension#getExtensionId() extensionId} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for extensionId (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtension withExtensionId(String value) {
    if (Objects.equals(this.extensionId, value)) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        value,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Extension#getStatus() status} attribute.
   * @param value The value for status
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withStatus(Extension.Status value) {
    Extension.Status newValue = Objects.requireNonNull(value, "status");
    if (this.status == newValue) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        newValue,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Extension#getStatus() status} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for status
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withStatus(Optional<? extends Extension.Status> optional) {
    Extension.Status value = optional.orElse(null);
    if (Objects.equals(this.status, value)) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        value,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Extension#getIcon() icon} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for icon (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtension withIcon(String value) {
    if (Objects.equals(this.icon, value)) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        value,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Extension#getDescription() description} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for description (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtension withDescription(String value) {
    if (Objects.equals(this.description, value)) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        value,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Extension#getUses() uses} attribute.
   * @param value The value for uses
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withUses(int value) {
    Integer newValue = value;
    if (Objects.equals(this.uses, newValue)) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        newValue,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Extension#getUses() uses} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for uses
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withUses(OptionalInt optional) {
    Integer value = optional.isPresent() ? optional.getAsInt() : null;
    if (Objects.equals(this.uses, value)) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        value,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Extension#getUserId() userId} attribute.
   * @param value The value for userId
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withUserId(String value) {
    String newValue = Objects.requireNonNull(value, "userId");
    if (Objects.equals(this.userId, newValue)) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        newValue,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Extension#getUserId() userId} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for userId
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withUserId(Optional<String> optional) {
    String value = optional.orElse(null);
    if (Objects.equals(this.userId, value)) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        value,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Extension#getLastUpdated() lastUpdated} attribute.
   * @param value The value for lastUpdated
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withLastUpdated(Date value) {
    Date newValue = Objects.requireNonNull(value, "lastUpdated");
    if (this.lastUpdated == newValue) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        newValue,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Extension#getLastUpdated() lastUpdated} attribute.
   * A shallow reference equality check is used on unboxed optional value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for lastUpdated
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withLastUpdated(Optional<? extends Date> optional) {
    Date value = optional.orElse(null);
    if (this.lastUpdated == value) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        value,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Extension#getCreatedDate() createdDate} attribute.
   * @param value The value for createdDate
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withCreatedDate(Date value) {
    Date newValue = Objects.requireNonNull(value, "createdDate");
    if (this.createdDate == newValue) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        newValue,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Extension#getCreatedDate() createdDate} attribute.
   * A shallow reference equality check is used on unboxed optional value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for createdDate
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withCreatedDate(Optional<? extends Date> optional) {
    Date value = optional.orElse(null);
    if (this.createdDate == value) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        value,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Extension#getId() id} attribute.
   * @param value The value for id
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withId(String value) {
    String newValue = Objects.requireNonNull(value, "id");
    if (Objects.equals(this.id, newValue)) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        newValue,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Extension#getId() id} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for id
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withId(Optional<String> optional) {
    String value = optional.orElse(null);
    if (Objects.equals(this.id, value)) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        value,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension#getActions() actions}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withActions(ExtensionAction... elements) {
    List<ExtensionAction> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        newValue,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension#getActions() actions}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of actions elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withActions(Iterable<? extends ExtensionAction> elements) {
    if (this.actions == elements) return this;
    List<ExtensionAction> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        newValue,
        this.name,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Extension#getName() name} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for name (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtension withName(String value) {
    if (Objects.equals(this.name, value)) return this;
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        value,
        this.tags,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension#getTags() tags}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withTags(String... elements) {
    SortedSet<String> newValue = createUnmodifiableSortedSet(false, createSafeList(Arrays.asList(elements), false, true));
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        newValue,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension#getTags() tags}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of tags elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withTags(Iterable<String> elements) {
    if (this.tags == elements) return this;
    SortedSet<String> newValue = createUnmodifiableSortedSet(false, createSafeList(elements, false, true));
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        newValue,
        this.properties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by replacing the {@link Extension#getProperties() properties} map with the specified map.
   * Nulls are not permitted as keys or values.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param entries The entries to be added to the properties map
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withProperties(Map<String, ? extends ConfigurationProperty> entries) {
    if (this.properties == entries) return this;
    Map<String, ConfigurationProperty> newValue = createUnmodifiableMap(true, false, entries);
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        newValue,
        this.dependencies));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension#getDependencies() dependencies}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withDependencies(Dependency... elements) {
    List<Dependency> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        newValue));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension#getDependencies() dependencies}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of dependencies elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withDependencies(Iterable<? extends Dependency> elements) {
    if (this.dependencies == elements) return this;
    List<Dependency> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableExtension(
        this,
        this.version,
        this.extensionId,
        this.status,
        this.icon,
        this.description,
        this.uses,
        this.userId,
        this.lastUpdated,
        this.createdDate,
        this.id,
        this.actions,
        this.name,
        this.tags,
        this.properties,
        newValue));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableExtension} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableExtension
        && equalTo((ImmutableExtension) another);
  }

  private boolean equalTo(ImmutableExtension another) {
    return Objects.equals(version, another.version)
        && Objects.equals(extensionId, another.extensionId)
        && Objects.equals(status, another.status)
        && Objects.equals(icon, another.icon)
        && Objects.equals(description, another.description)
        && Objects.equals(uses, another.uses)
        && Objects.equals(userId, another.userId)
        && Objects.equals(lastUpdated, another.lastUpdated)
        && Objects.equals(createdDate, another.createdDate)
        && Objects.equals(id, another.id)
        && actions.equals(another.actions)
        && Objects.equals(name, another.name)
        && tags.equals(another.tags)
        && properties.equals(another.properties)
        && dependencies.equals(another.dependencies);
  }

  /**
   * Computes a hash code from attributes: {@code version}, {@code extensionId}, {@code status}, {@code icon}, {@code description}, {@code uses}, {@code userId}, {@code lastUpdated}, {@code createdDate}, {@code id}, {@code actions}, {@code name}, {@code tags}, {@code properties}, {@code dependencies}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + Objects.hashCode(version);
    h += (h << 5) + Objects.hashCode(extensionId);
    h += (h << 5) + Objects.hashCode(status);
    h += (h << 5) + Objects.hashCode(icon);
    h += (h << 5) + Objects.hashCode(description);
    h += (h << 5) + Objects.hashCode(uses);
    h += (h << 5) + Objects.hashCode(userId);
    h += (h << 5) + Objects.hashCode(lastUpdated);
    h += (h << 5) + Objects.hashCode(createdDate);
    h += (h << 5) + Objects.hashCode(id);
    h += (h << 5) + actions.hashCode();
    h += (h << 5) + Objects.hashCode(name);
    h += (h << 5) + tags.hashCode();
    h += (h << 5) + properties.hashCode();
    h += (h << 5) + dependencies.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code Extension} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder("Extension{");
    if (version != null) {
      builder.append("version=").append(version);
    }
    if (extensionId != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("extensionId=").append(extensionId);
    }
    if (status != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("status=").append(status);
    }
    if (icon != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("icon=").append(icon);
    }
    if (description != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("description=").append(description);
    }
    if (uses != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("uses=").append(uses);
    }
    if (userId != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("userId=").append(userId);
    }
    if (lastUpdated != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("lastUpdated=").append(lastUpdated);
    }
    if (createdDate != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("createdDate=").append(createdDate);
    }
    if (id != null) {
      if (builder.length() > 10) builder.append(", ");
      builder.append("id=").append(id);
    }
    if (builder.length() > 10) builder.append(", ");
    builder.append("actions=").append(actions);
    if (name != null) {
      builder.append(", ");
      builder.append("name=").append(name);
    }
    builder.append(", ");
    builder.append("tags=").append(tags);
    builder.append(", ");
    builder.append("properties=").append(properties);
    builder.append(", ");
    builder.append("dependencies=").append(dependencies);
    return builder.append("}").toString();
  }

  /**
   * Construct a new immutable {@code Extension} instance.
   * @param version The value for the {@code version} attribute
   * @param extensionId The value for the {@code extensionId} attribute
   * @param status The value for the {@code status} attribute
   * @param icon The value for the {@code icon} attribute
   * @param description The value for the {@code description} attribute
   * @param uses The value for the {@code uses} attribute
   * @param userId The value for the {@code userId} attribute
   * @param lastUpdated The value for the {@code lastUpdated} attribute
   * @param createdDate The value for the {@code createdDate} attribute
   * @param id The value for the {@code id} attribute
   * @param actions The value for the {@code actions} attribute
   * @param name The value for the {@code name} attribute
   * @param tags The value for the {@code tags} attribute
   * @param properties The value for the {@code properties} attribute
   * @param dependencies The value for the {@code dependencies} attribute
   * @return An immutable Extension instance
   */
  public static Extension of(String version, String extensionId, Optional<Extension.Status> status, String icon, String description, OptionalInt uses, Optional<String> userId, Optional<Date> lastUpdated, Optional<Date> createdDate, Optional<String> id, List<ExtensionAction> actions, String name, SortedSet<String> tags, Map<String, ConfigurationProperty> properties, List<Dependency> dependencies) {
    return of(version, extensionId, status, icon, description, uses, userId, lastUpdated, createdDate, id, (Iterable<? extends ExtensionAction>) actions, name, (Iterable<String>) tags, properties, (Iterable<? extends Dependency>) dependencies);
  }

  /**
   * Construct a new immutable {@code Extension} instance.
   * @param version The value for the {@code version} attribute
   * @param extensionId The value for the {@code extensionId} attribute
   * @param status The value for the {@code status} attribute
   * @param icon The value for the {@code icon} attribute
   * @param description The value for the {@code description} attribute
   * @param uses The value for the {@code uses} attribute
   * @param userId The value for the {@code userId} attribute
   * @param lastUpdated The value for the {@code lastUpdated} attribute
   * @param createdDate The value for the {@code createdDate} attribute
   * @param id The value for the {@code id} attribute
   * @param actions The value for the {@code actions} attribute
   * @param name The value for the {@code name} attribute
   * @param tags The value for the {@code tags} attribute
   * @param properties The value for the {@code properties} attribute
   * @param dependencies The value for the {@code dependencies} attribute
   * @return An immutable Extension instance
   */
  public static Extension of(String version, String extensionId, Optional<Extension.Status> status, String icon, String description, OptionalInt uses, Optional<String> userId, Optional<Date> lastUpdated, Optional<Date> createdDate, Optional<String> id, Iterable<? extends ExtensionAction> actions, String name, Iterable<String> tags, Map<String, ? extends ConfigurationProperty> properties, Iterable<? extends Dependency> dependencies) {
    return validate(new ImmutableExtension(version, extensionId, status, icon, description, uses, userId, lastUpdated, createdDate, id, actions, name, tags, properties, dependencies));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableExtension validate(ImmutableExtension instance) {
    Set<ConstraintViolation<ImmutableExtension>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link Extension} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Extension instance
   */
  public static Extension copyOf(Extension instance) {
    if (instance instanceof ImmutableExtension) {
      return (ImmutableExtension) instance;
    }
    return new Extension.Builder()
        .createFrom(instance)
        .build();
  }

  private Object readResolve() throws ObjectStreamException {
    return validate(this);
  }

  /**
   * Builds instances of type {@link Extension Extension}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  @JsonPropertyOrder({"name", "description", "icon", "extensionId", "version", "tags", "actions", "dependencies"})
  public static class Builder {
    private static final long OPT_BIT_DEPENDENCIES = 0x1L;
    private long optBits;

    private String version;
    private String extensionId;
    private Extension.Status status;
    private String icon;
    private String description;
    private Integer uses;
    private String userId;
    private Date lastUpdated;
    private Date createdDate;
    private String id;
    private List<ExtensionAction> actions = new ArrayList<ExtensionAction>();
    private String name;
    private List<String> tags = new ArrayList<String>();
    private Map<String, ConfigurationProperty> properties = new LinkedHashMap<String, ConfigurationProperty>();
    private List<Dependency> dependencies = new ArrayList<Dependency>();

    /**
     * Creates a builder for {@link Extension Extension} instances.
     */
    public Builder() {
      if (!(this instanceof Extension.Builder)) {
        throw new UnsupportedOperationException("Use: new Extension.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithConfigurationProperties} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder createFrom(WithConfigurationProperties instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (Extension.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.extension.Extension} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder createFrom(Extension instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (Extension.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithDependencies} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder createFrom(WithDependencies instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (Extension.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithTags} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder createFrom(WithTags instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (Extension.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithName} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder createFrom(WithName instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (Extension.Builder) this;
    }

    private void from(Object object) {
      if (object instanceof WithConfigurationProperties) {
        WithConfigurationProperties instance = (WithConfigurationProperties) object;
        putAllProperties(instance.getProperties());
      }
      if (object instanceof Extension) {
        Extension instance = (Extension) object;
        Optional<Date> lastUpdatedOptional = instance.getLastUpdated();
        if (lastUpdatedOptional.isPresent()) {
          lastUpdated(lastUpdatedOptional);
        }
        Optional<Date> createdDateOptional = instance.getCreatedDate();
        if (createdDateOptional.isPresent()) {
          createdDate(createdDateOptional);
        }
        String iconValue = instance.getIcon();
        if (iconValue != null) {
          icon(iconValue);
        }
        String descriptionValue = instance.getDescription();
        if (descriptionValue != null) {
          description(descriptionValue);
        }
        OptionalInt usesOptional = instance.getUses();
        if (usesOptional.isPresent()) {
          uses(usesOptional);
        }
        Optional<String> idOptional = instance.getId();
        if (idOptional.isPresent()) {
          id(idOptional);
        }
        String versionValue = instance.getVersion();
        if (versionValue != null) {
          version(versionValue);
        }
        Optional<String> userIdOptional = instance.getUserId();
        if (userIdOptional.isPresent()) {
          userId(userIdOptional);
        }
        String extensionIdValue = instance.getExtensionId();
        if (extensionIdValue != null) {
          extensionId(extensionIdValue);
        }
        addAllActions(instance.getActions());
        Optional<Extension.Status> statusOptional = instance.getStatus();
        if (statusOptional.isPresent()) {
          status(statusOptional);
        }
      }
      if (object instanceof WithDependencies) {
        WithDependencies instance = (WithDependencies) object;
        addAllDependencies(instance.getDependencies());
      }
      if (object instanceof WithTags) {
        WithTags instance = (WithTags) object;
        addAllTags(instance.getTags());
      }
      if (object instanceof WithName) {
        WithName instance = (WithName) object;
        String nameValue = instance.getName();
        if (nameValue != null) {
          name(nameValue);
        }
      }
    }

    /**
     * Initializes the value for the {@link Extension#getVersion() version} attribute.
     * @param version The value for version (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("version")
    public final Extension.Builder version(String version) {
      this.version = version;
      return (Extension.Builder) this;
    }

    /**
     * Initializes the value for the {@link Extension#getExtensionId() extensionId} attribute.
     * @param extensionId The value for extensionId (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("extensionId")
    public final Extension.Builder extensionId(String extensionId) {
      this.extensionId = extensionId;
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getStatus() status} to status.
     * @param status The value for status
     * @return {@code this} builder for chained invocation
     */
    public final Extension.Builder status(Extension.Status status) {
      this.status = Objects.requireNonNull(status, "status");
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getStatus() status} to status.
     * @param status The value for status
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("status")
    public final Extension.Builder status(Optional<? extends Extension.Status> status) {
      this.status = status.orElse(null);
      return (Extension.Builder) this;
    }

    /**
     * Initializes the value for the {@link Extension#getIcon() icon} attribute.
     * @param icon The value for icon (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("icon")
    public final Extension.Builder icon(String icon) {
      this.icon = icon;
      return (Extension.Builder) this;
    }

    /**
     * Initializes the value for the {@link Extension#getDescription() description} attribute.
     * @param description The value for description (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("description")
    public final Extension.Builder description(String description) {
      this.description = description;
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getUses() uses} to uses.
     * @param uses The value for uses
     * @return {@code this} builder for chained invocation
     */
    public final Extension.Builder uses(int uses) {
      this.uses = uses;
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getUses() uses} to uses.
     * @param uses The value for uses
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("uses")
    public final Extension.Builder uses(OptionalInt uses) {
      this.uses = uses.isPresent() ? uses.getAsInt() : null;
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getUserId() userId} to userId.
     * @param userId The value for userId
     * @return {@code this} builder for chained invocation
     */
    public final Extension.Builder userId(String userId) {
      this.userId = Objects.requireNonNull(userId, "userId");
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getUserId() userId} to userId.
     * @param userId The value for userId
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("userId")
    public final Extension.Builder userId(Optional<String> userId) {
      this.userId = userId.orElse(null);
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getLastUpdated() lastUpdated} to lastUpdated.
     * @param lastUpdated The value for lastUpdated
     * @return {@code this} builder for chained invocation
     */
    public final Extension.Builder lastUpdated(Date lastUpdated) {
      this.lastUpdated = Objects.requireNonNull(lastUpdated, "lastUpdated");
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getLastUpdated() lastUpdated} to lastUpdated.
     * @param lastUpdated The value for lastUpdated
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("lastUpdated")
    public final Extension.Builder lastUpdated(Optional<? extends Date> lastUpdated) {
      this.lastUpdated = lastUpdated.orElse(null);
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getCreatedDate() createdDate} to createdDate.
     * @param createdDate The value for createdDate
     * @return {@code this} builder for chained invocation
     */
    public final Extension.Builder createdDate(Date createdDate) {
      this.createdDate = Objects.requireNonNull(createdDate, "createdDate");
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getCreatedDate() createdDate} to createdDate.
     * @param createdDate The value for createdDate
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("createdDate")
    public final Extension.Builder createdDate(Optional<? extends Date> createdDate) {
      this.createdDate = createdDate.orElse(null);
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getId() id} to id.
     * @param id The value for id
     * @return {@code this} builder for chained invocation
     */
    public final Extension.Builder id(String id) {
      this.id = Objects.requireNonNull(id, "id");
      return (Extension.Builder) this;
    }

    /**
     * Initializes the optional value {@link Extension#getId() id} to id.
     * @param id The value for id
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("id")
    public final Extension.Builder id(Optional<String> id) {
      this.id = id.orElse(null);
      return (Extension.Builder) this;
    }

    /**
     * Adds one element to {@link Extension#getActions() actions} list.
     * @param element A actions element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder addAction(ExtensionAction element) {
      this.actions.add(Objects.requireNonNull(element, "actions element"));
      return (Extension.Builder) this;
    }

    /**
     * Adds elements to {@link Extension#getActions() actions} list.
     * @param elements An array of actions elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder addAction(ExtensionAction... elements) {
      for (ExtensionAction element : elements) {
        this.actions.add(Objects.requireNonNull(element, "actions element"));
      }
      return (Extension.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link Extension#getActions() actions} list.
     * @param elements An iterable of actions elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("actions")
    public final Extension.Builder actions(Iterable<? extends ExtensionAction> elements) {
      this.actions.clear();
      return addAllActions(elements);
    }

    /**
     * Adds elements to {@link Extension#getActions() actions} list.
     * @param elements An iterable of actions elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder addAllActions(Iterable<? extends ExtensionAction> elements) {
      for (ExtensionAction element : elements) {
        this.actions.add(Objects.requireNonNull(element, "actions element"));
      }
      return (Extension.Builder) this;
    }

    /**
     * Initializes the value for the {@link Extension#getName() name} attribute.
     * @param name The value for name (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("name")
    public final Extension.Builder name(String name) {
      this.name = name;
      return (Extension.Builder) this;
    }

    /**
     * Adds one element to {@link Extension#getTags() tags} sortedSet.
     * @param element A tags element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder addTag(String element) {
      if (element != null) this.tags.add(element);
      return (Extension.Builder) this;
    }

    /**
     * Adds elements to {@link Extension#getTags() tags} sortedSet.
     * @param elements An array of tags elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder addTag(String... elements) {
      for (String element : elements) {
        if (element != null) this.tags.add(element);
      }
      return (Extension.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link Extension#getTags() tags} sortedSet.
     * @param elements An iterable of tags elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("tags")
    public final Extension.Builder tags(Iterable<String> elements) {
      this.tags.clear();
      return addAllTags(elements);
    }

    /**
     * Adds elements to {@link Extension#getTags() tags} sortedSet.
     * @param elements An iterable of tags elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder addAllTags(Iterable<String> elements) {
      for (String element : elements) {
        if (element != null) this.tags.add(element);
      }
      return (Extension.Builder) this;
    }

    /**
     * Put one entry to the {@link Extension#getProperties() properties} map.
     * @param key The key in the properties map
     * @param value The associated value in the properties map
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder putProperty(String key, ConfigurationProperty value) {
      this.properties.put(
          Objects.requireNonNull(key, "properties key"),
          Objects.requireNonNull(value, "properties value"));
      return (Extension.Builder) this;
    }

    /**
     * Put one entry to the {@link Extension#getProperties() properties} map. Nulls are not permitted
     * @param entry The key and value entry
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder putProperty(Map.Entry<String, ? extends ConfigurationProperty> entry) {
      String k = entry.getKey();
      ConfigurationProperty v = entry.getValue();
      this.properties.put(
          Objects.requireNonNull(k, "properties key"),
          Objects.requireNonNull(v, "properties value"));
      return (Extension.Builder) this;
    }

    /**
     * Sets or replaces all mappings from the specified map as entries for the {@link Extension#getProperties() properties} map. Nulls are not permitted
     * @param properties The entries that will be added to the properties map
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("properties")
    public final Extension.Builder properties(Map<String, ? extends ConfigurationProperty> properties) {
      this.properties.clear();
      return putAllProperties(properties);
    }

    /**
     * Put all mappings from the specified map as entries to {@link Extension#getProperties() properties} map. Nulls are not permitted
     * @param properties The entries that will be added to the properties map
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder putAllProperties(Map<String, ? extends ConfigurationProperty> properties) {
      for (Map.Entry<String, ? extends ConfigurationProperty> entry : properties.entrySet()) {
        String k = entry.getKey();
        ConfigurationProperty v = entry.getValue();
        this.properties.put(
            Objects.requireNonNull(k, "properties key"),
            Objects.requireNonNull(v, "properties value"));
      }
      return (Extension.Builder) this;
    }

    /**
     * Adds one element to {@link Extension#getDependencies() dependencies} list.
     * @param element A dependencies element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder addDependency(Dependency element) {
      this.dependencies.add(Objects.requireNonNull(element, "dependencies element"));
      optBits |= OPT_BIT_DEPENDENCIES;
      return (Extension.Builder) this;
    }

    /**
     * Adds elements to {@link Extension#getDependencies() dependencies} list.
     * @param elements An array of dependencies elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder addDependency(Dependency... elements) {
      for (Dependency element : elements) {
        this.dependencies.add(Objects.requireNonNull(element, "dependencies element"));
      }
      optBits |= OPT_BIT_DEPENDENCIES;
      return (Extension.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link Extension#getDependencies() dependencies} list.
     * @param elements An iterable of dependencies elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("dependencies")
    public final Extension.Builder dependencies(Iterable<? extends Dependency> elements) {
      this.dependencies.clear();
      return addAllDependencies(elements);
    }

    /**
     * Adds elements to {@link Extension#getDependencies() dependencies} list.
     * @param elements An iterable of dependencies elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Extension.Builder addAllDependencies(Iterable<? extends Dependency> elements) {
      for (Dependency element : elements) {
        this.dependencies.add(Objects.requireNonNull(element, "dependencies element"));
      }
      optBits |= OPT_BIT_DEPENDENCIES;
      return (Extension.Builder) this;
    }

    /**
     * Builds a new {@link Extension Extension}.
     * @return An immutable instance of Extension
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public Extension build() {
      return ImmutableExtension.validate(new ImmutableExtension(this));
    }

    private boolean dependenciesIsSet() {
      return (optBits & OPT_BIT_DEPENDENCIES) != 0;
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<T>();
    } else {
      list = new ArrayList<T>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<T>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }

  private static <T extends Comparable<T>> NavigableSet<T> createUnmodifiableSortedSet(boolean reverse, List<T> list) {
    TreeSet<T> set = reverse
        ? new TreeSet<T>(Collections.reverseOrder())
        : new TreeSet<T>();
    set.addAll(list);
    return Collections.unmodifiableNavigableSet(set);
  }

  private static <K, V> Map<K, V> createUnmodifiableMap(boolean checkNulls, boolean skipNulls, Map<? extends K, ? extends V> map) {
    switch (map.size()) {
    case 0: return Collections.emptyMap();
    case 1: {
      Map.Entry<? extends K, ? extends V> e = map.entrySet().iterator().next();
      K k = e.getKey();
      V v = e.getValue();
      if (checkNulls) {
        Objects.requireNonNull(k, "key");
        Objects.requireNonNull(v, "value");
      }
      if (skipNulls && (k == null || v == null)) {
        return Collections.emptyMap();
      }
      return Collections.singletonMap(k, v);
    }
    default: {
      Map<K, V> linkedMap = new LinkedHashMap<K, V>(map.size());
      if (skipNulls || checkNulls) {
        for (Map.Entry<? extends K, ? extends V> e : map.entrySet()) {
          K k = e.getKey();
          V v = e.getValue();
          if (skipNulls) {
            if (k == null || v == null) continue;
          } else if (checkNulls) {
            Objects.requireNonNull(k, "key");
            Objects.requireNonNull(v, "value");
          }
          linkedMap.put(k, v);
        }
      } else {
        linkedMap.putAll(map);
      }
      return Collections.unmodifiableMap(linkedMap);
    }
    }
  }
}
