/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.model.filter;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonProperty;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import io.syndesis.model.Dependency;
import io.syndesis.model.Kind;
import io.syndesis.model.WithConfiguredProperties;
import io.syndesis.model.WithDependencies;
import io.syndesis.model.WithKind;
import io.syndesis.model.action.Action;
import io.syndesis.model.connection.Connection;
import io.syndesis.model.extension.Extension;
import io.syndesis.model.integration.Step;
import java.io.ObjectStreamException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import javax.annotation.Generated;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;

/**
 * Immutable implementation of {@link ExpressionFilterStep}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new ExpressionFilterStep.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableExpressionFilterStep.of()}.
 */
@SuppressWarnings({"all"})
@SuppressFBWarnings
@Generated({"Immutables.generator", "ExpressionFilterStep"})
@JsonIgnoreProperties({"filterExpression"})
final class ImmutableExpressionFilterStep implements ExpressionFilterStep {
  private final String stepKind;
  private final Kind kind;
  private final Action<?> action;
  private final Connection connection;
  private final Extension extension;
  private final String name;
  private final String id;
  private final Map<String, String> configuredProperties;
  private final List<Dependency> dependencies;

  private ImmutableExpressionFilterStep(
      String stepKind,
      Kind kind,
      Optional<Action<?>> action,
      Optional<Connection> connection,
      Optional<Extension> extension,
      String name,
      Optional<String> id,
      Map<String, ? extends String> configuredProperties,
      Iterable<? extends Dependency> dependencies) {
    this.stepKind = Objects.requireNonNull(stepKind, "stepKind");
    this.kind = Objects.requireNonNull(kind, "kind");
    this.action = action.orElse(null);
    this.connection = connection.orElse(null);
    this.extension = extension.orElse(null);
    this.name = name;
    this.id = id.orElse(null);
    this.configuredProperties = createUnmodifiableMap(true, false, configuredProperties);
    this.dependencies = createUnmodifiableList(false, createSafeList(dependencies, true, false));
    initShim.stepKind(this.stepKind);
    initShim.kind(this.kind);
    initShim.configuredProperties(this.configuredProperties);
    initShim.dependencies(this.dependencies);
    this.initShim = null;
  }

  private ImmutableExpressionFilterStep(ImmutableExpressionFilterStep.Builder builder) {
    this.action = builder.action;
    this.connection = builder.connection;
    this.extension = builder.extension;
    this.name = builder.name;
    this.id = builder.id;
    if (builder.stepKind != null) {
      initShim.stepKind(builder.stepKind);
    }
    if (builder.kind != null) {
      initShim.kind(builder.kind);
    }
    if (builder.configuredPropertiesIsSet()) {
      initShim.configuredProperties(createUnmodifiableMap(false, false, builder.configuredProperties));
    }
    if (builder.dependenciesIsSet()) {
      initShim.dependencies(createUnmodifiableList(true, builder.dependencies));
    }
    this.stepKind = initShim.getStepKind();
    this.kind = initShim.getKind();
    this.configuredProperties = initShim.getConfiguredProperties();
    this.dependencies = initShim.getDependencies();
    this.initShim = null;
  }

  private ImmutableExpressionFilterStep(
      ImmutableExpressionFilterStep original,
      String stepKind,
      Kind kind,
      Action<?> action,
      Connection connection,
      Extension extension,
      String name,
      String id,
      Map<String, String> configuredProperties,
      List<Dependency> dependencies) {
    this.stepKind = stepKind;
    this.kind = kind;
    this.action = action;
    this.connection = connection;
    this.extension = extension;
    this.name = name;
    this.id = id;
    this.configuredProperties = configuredProperties;
    this.dependencies = dependencies;
    this.initShim = null;
  }

  private static final int STAGE_INITIALIZING = -1;
  private static final int STAGE_UNINITIALIZED = 0;
  private static final int STAGE_INITIALIZED = 1;
  private transient volatile InitShim initShim = new InitShim();

  private final class InitShim {
    private String stepKind;
    private int stepKindBuildStage;

    String getStepKind() {
      if (stepKindBuildStage == STAGE_INITIALIZING) throw new IllegalStateException(formatInitCycleMessage());
      if (stepKindBuildStage == STAGE_UNINITIALIZED) {
        stepKindBuildStage = STAGE_INITIALIZING;
        this.stepKind = Objects.requireNonNull(getStepKindInitialize(), "stepKind");
        stepKindBuildStage = STAGE_INITIALIZED;
      }
      return this.stepKind;
    }

    void stepKind(String stepKind) {
      this.stepKind = stepKind;
      stepKindBuildStage = STAGE_INITIALIZED;
    }
    private Kind kind;
    private int kindBuildStage;

    Kind getKind() {
      if (kindBuildStage == STAGE_INITIALIZING) throw new IllegalStateException(formatInitCycleMessage());
      if (kindBuildStage == STAGE_UNINITIALIZED) {
        kindBuildStage = STAGE_INITIALIZING;
        this.kind = Objects.requireNonNull(getKindInitialize(), "kind");
        kindBuildStage = STAGE_INITIALIZED;
      }
      return this.kind;
    }

    void kind(Kind kind) {
      this.kind = kind;
      kindBuildStage = STAGE_INITIALIZED;
    }
    private Map<String, String> configuredProperties;
    private int configuredPropertiesBuildStage;

    Map<String, String> getConfiguredProperties() {
      if (configuredPropertiesBuildStage == STAGE_INITIALIZING) throw new IllegalStateException(formatInitCycleMessage());
      if (configuredPropertiesBuildStage == STAGE_UNINITIALIZED) {
        configuredPropertiesBuildStage = STAGE_INITIALIZING;
        this.configuredProperties = createUnmodifiableMap(true, false, getConfiguredPropertiesInitialize());
        configuredPropertiesBuildStage = STAGE_INITIALIZED;
      }
      return this.configuredProperties;
    }

    void configuredProperties(Map<String, String> configuredProperties) {
      this.configuredProperties = configuredProperties;
      configuredPropertiesBuildStage = STAGE_INITIALIZED;
    }
    private List<Dependency> dependencies;
    private int dependenciesBuildStage;

    List<Dependency> getDependencies() {
      if (dependenciesBuildStage == STAGE_INITIALIZING) throw new IllegalStateException(formatInitCycleMessage());
      if (dependenciesBuildStage == STAGE_UNINITIALIZED) {
        dependenciesBuildStage = STAGE_INITIALIZING;
        this.dependencies = createUnmodifiableList(false, createSafeList(getDependenciesInitialize(), true, false));
        dependenciesBuildStage = STAGE_INITIALIZED;
      }
      return this.dependencies;
    }

    void dependencies(List<Dependency> dependencies) {
      this.dependencies = dependencies;
      dependenciesBuildStage = STAGE_INITIALIZED;
    }

    private String formatInitCycleMessage() {
      ArrayList<String> attributes = new ArrayList<String>();
      if (stepKindBuildStage == STAGE_INITIALIZING) attributes.add("stepKind");
      if (kindBuildStage == STAGE_INITIALIZING) attributes.add("kind");
      if (configuredPropertiesBuildStage == STAGE_INITIALIZING) attributes.add("configuredProperties");
      if (dependenciesBuildStage == STAGE_INITIALIZING) attributes.add("dependencies");
      return "Cannot build ExpressionFilterStep, attribute initializers form cycle" + attributes;
    }
  }

  private String getStepKindInitialize() {
    return ExpressionFilterStep.super.getStepKind();
  }

  private Kind getKindInitialize() {
    return ExpressionFilterStep.super.getKind();
  }

  private Map<String, String> getConfiguredPropertiesInitialize() {
    return ExpressionFilterStep.super.getConfiguredProperties();
  }

  private List<Dependency> getDependenciesInitialize() {
    return ExpressionFilterStep.super.getDependencies();
  }

  /**
   * @return The value of the {@code stepKind} attribute
   */
  @JsonProperty("stepKind")
  @Override
  public String getStepKind() {
    InitShim shim = this.initShim;
    return shim != null
        ? shim.getStepKind()
        : this.stepKind;
  }

  /**
   * @return The value of the {@code kind} attribute
   */
  @JsonProperty("kind")
  @Override
  public Kind getKind() {
    InitShim shim = this.initShim;
    return shim != null
        ? shim.getKind()
        : this.kind;
  }

  /**
   * @return The value of the {@code action} attribute
   */
  @JsonProperty("action")
  @Override
  public Optional<Action<?>> getAction() {
    return Optional.ofNullable(action);
  }

  /**
   * @return The value of the {@code connection} attribute
   */
  @JsonProperty("connection")
  @Override
  public Optional<Connection> getConnection() {
    return Optional.ofNullable(connection);
  }

  /**
   * @return The value of the {@code extension} attribute
   */
  @JsonProperty("extension")
  @Override
  public Optional<Extension> getExtension() {
    return Optional.ofNullable(extension);
  }

  /**
   * @return The value of the {@code name} attribute
   */
  @JsonProperty("name")
  @Override
  public String getName() {
    return name;
  }

  /**
   * @return The value of the {@code id} attribute
   */
  @JsonProperty("id")
  @Override
  public Optional<String> getId() {
    return Optional.ofNullable(id);
  }

  /**
   * @return The value of the {@code configuredProperties} attribute
   */
  @JsonProperty("configuredProperties")
  @Override
  public Map<String, String> getConfiguredProperties() {
    InitShim shim = this.initShim;
    return shim != null
        ? shim.getConfiguredProperties()
        : this.configuredProperties;
  }

  /**
   * @return The value of the {@code dependencies} attribute
   */
  @JsonProperty("dependencies")
  @Override
  public List<Dependency> getDependencies() {
    InitShim shim = this.initShim;
    return shim != null
        ? shim.getDependencies()
        : this.dependencies;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ExpressionFilterStep#getStepKind() stepKind} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for stepKind
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExpressionFilterStep withStepKind(String value) {
    if (this.stepKind.equals(value)) return this;
    String newValue = Objects.requireNonNull(value, "stepKind");
    return validate(new ImmutableExpressionFilterStep(
        this,
        newValue,
        this.kind,
        this.action,
        this.connection,
        this.extension,
        this.name,
        this.id,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ExpressionFilterStep#getKind() kind} attribute.
   * A value equality check is used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for kind
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExpressionFilterStep withKind(Kind value) {
    if (this.kind == value) return this;
    Kind newValue = Objects.requireNonNull(value, "kind");
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        newValue,
        this.action,
        this.connection,
        this.extension,
        this.name,
        this.id,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link ExpressionFilterStep#getAction() action} attribute.
   * @param value The value for action
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExpressionFilterStep withAction(Action<?> value) {
    Action<?> newValue = Objects.requireNonNull(value, "action");
    if (this.action == newValue) return this;
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        newValue,
        this.connection,
        this.extension,
        this.name,
        this.id,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link ExpressionFilterStep#getAction() action} attribute.
   * A shallow reference equality check is used on unboxed optional value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for action
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExpressionFilterStep withAction(Optional<? extends Action<?>> optional) {
    Action<?> value = optional.orElse(null);
    if (this.action == value) return this;
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        value,
        this.connection,
        this.extension,
        this.name,
        this.id,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link ExpressionFilterStep#getConnection() connection} attribute.
   * @param value The value for connection
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExpressionFilterStep withConnection(Connection value) {
    Connection newValue = Objects.requireNonNull(value, "connection");
    if (this.connection == newValue) return this;
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        this.action,
        newValue,
        this.extension,
        this.name,
        this.id,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link ExpressionFilterStep#getConnection() connection} attribute.
   * A shallow reference equality check is used on unboxed optional value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for connection
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExpressionFilterStep withConnection(Optional<? extends Connection> optional) {
    Connection value = optional.orElse(null);
    if (this.connection == value) return this;
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        this.action,
        value,
        this.extension,
        this.name,
        this.id,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link ExpressionFilterStep#getExtension() extension} attribute.
   * @param value The value for extension
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExpressionFilterStep withExtension(Extension value) {
    Extension newValue = Objects.requireNonNull(value, "extension");
    if (this.extension == newValue) return this;
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        this.action,
        this.connection,
        newValue,
        this.name,
        this.id,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link ExpressionFilterStep#getExtension() extension} attribute.
   * A shallow reference equality check is used on unboxed optional value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for extension
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExpressionFilterStep withExtension(Optional<? extends Extension> optional) {
    Extension value = optional.orElse(null);
    if (this.extension == value) return this;
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        this.action,
        this.connection,
        value,
        this.name,
        this.id,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ExpressionFilterStep#getName() name} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for name (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExpressionFilterStep withName(String value) {
    if (Objects.equals(this.name, value)) return this;
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        this.action,
        this.connection,
        this.extension,
        value,
        this.id,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link ExpressionFilterStep#getId() id} attribute.
   * @param value The value for id
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExpressionFilterStep withId(String value) {
    String newValue = Objects.requireNonNull(value, "id");
    if (Objects.equals(this.id, newValue)) return this;
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        this.action,
        this.connection,
        this.extension,
        this.name,
        newValue,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link ExpressionFilterStep#getId() id} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for id
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExpressionFilterStep withId(Optional<String> optional) {
    String value = optional.orElse(null);
    if (Objects.equals(this.id, value)) return this;
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        this.action,
        this.connection,
        this.extension,
        this.name,
        value,
        this.configuredProperties,
        this.dependencies));
  }

  /**
   * Copy the current immutable object by replacing the {@link ExpressionFilterStep#getConfiguredProperties() configuredProperties} map with the specified map.
   * Nulls are not permitted as keys or values.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param entries The entries to be added to the configuredProperties map
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExpressionFilterStep withConfiguredProperties(Map<String, ? extends String> entries) {
    if (this.configuredProperties == entries) return this;
    Map<String, String> newValue = createUnmodifiableMap(true, false, entries);
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        this.action,
        this.connection,
        this.extension,
        this.name,
        this.id,
        newValue,
        this.dependencies));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ExpressionFilterStep#getDependencies() dependencies}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExpressionFilterStep withDependencies(Dependency... elements) {
    List<Dependency> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        this.action,
        this.connection,
        this.extension,
        this.name,
        this.id,
        this.configuredProperties,
        newValue));
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link ExpressionFilterStep#getDependencies() dependencies}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of dependencies elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExpressionFilterStep withDependencies(Iterable<? extends Dependency> elements) {
    if (this.dependencies == elements) return this;
    List<Dependency> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return validate(new ImmutableExpressionFilterStep(
        this,
        this.stepKind,
        this.kind,
        this.action,
        this.connection,
        this.extension,
        this.name,
        this.id,
        this.configuredProperties,
        newValue));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableExpressionFilterStep} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableExpressionFilterStep
        && equalTo((ImmutableExpressionFilterStep) another);
  }

  private boolean equalTo(ImmutableExpressionFilterStep another) {
    return stepKind.equals(another.stepKind)
        && kind.equals(another.kind)
        && Objects.equals(action, another.action)
        && Objects.equals(connection, another.connection)
        && Objects.equals(extension, another.extension)
        && Objects.equals(name, another.name)
        && Objects.equals(id, another.id)
        && configuredProperties.equals(another.configuredProperties)
        && dependencies.equals(another.dependencies);
  }

  /**
   * Computes a hash code from attributes: {@code stepKind}, {@code kind}, {@code action}, {@code connection}, {@code extension}, {@code name}, {@code id}, {@code configuredProperties}, {@code dependencies}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + stepKind.hashCode();
    h += (h << 5) + kind.hashCode();
    h += (h << 5) + Objects.hashCode(action);
    h += (h << 5) + Objects.hashCode(connection);
    h += (h << 5) + Objects.hashCode(extension);
    h += (h << 5) + Objects.hashCode(name);
    h += (h << 5) + Objects.hashCode(id);
    h += (h << 5) + configuredProperties.hashCode();
    h += (h << 5) + dependencies.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code ExpressionFilterStep} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder("ExpressionFilterStep{");
    builder.append("stepKind=").append(stepKind);
    builder.append(", ");
    builder.append("kind=").append(kind);
    if (action != null) {
      builder.append(", ");
      builder.append("action=").append(action);
    }
    if (connection != null) {
      builder.append(", ");
      builder.append("connection=").append(connection);
    }
    if (extension != null) {
      builder.append(", ");
      builder.append("extension=").append(extension);
    }
    if (name != null) {
      builder.append(", ");
      builder.append("name=").append(name);
    }
    if (id != null) {
      builder.append(", ");
      builder.append("id=").append(id);
    }
    builder.append(", ");
    builder.append("configuredProperties=").append(configuredProperties);
    builder.append(", ");
    builder.append("dependencies=").append(dependencies);
    return builder.append("}").toString();
  }

  /**
   * Construct a new immutable {@code ExpressionFilterStep} instance.
   * @param stepKind The value for the {@code stepKind} attribute
   * @param kind The value for the {@code kind} attribute
   * @param action The value for the {@code action} attribute
   * @param connection The value for the {@code connection} attribute
   * @param extension The value for the {@code extension} attribute
   * @param name The value for the {@code name} attribute
   * @param id The value for the {@code id} attribute
   * @param configuredProperties The value for the {@code configuredProperties} attribute
   * @param dependencies The value for the {@code dependencies} attribute
   * @return An immutable ExpressionFilterStep instance
   */
  public static ExpressionFilterStep of(String stepKind, Kind kind, Optional<Action<?>> action, Optional<Connection> connection, Optional<Extension> extension, String name, Optional<String> id, Map<String, String> configuredProperties, List<Dependency> dependencies) {
    return of(stepKind, kind, action, connection, extension, name, id, configuredProperties, (Iterable<? extends Dependency>) dependencies);
  }

  /**
   * Construct a new immutable {@code ExpressionFilterStep} instance.
   * @param stepKind The value for the {@code stepKind} attribute
   * @param kind The value for the {@code kind} attribute
   * @param action The value for the {@code action} attribute
   * @param connection The value for the {@code connection} attribute
   * @param extension The value for the {@code extension} attribute
   * @param name The value for the {@code name} attribute
   * @param id The value for the {@code id} attribute
   * @param configuredProperties The value for the {@code configuredProperties} attribute
   * @param dependencies The value for the {@code dependencies} attribute
   * @return An immutable ExpressionFilterStep instance
   */
  public static ExpressionFilterStep of(String stepKind, Kind kind, Optional<Action<?>> action, Optional<Connection> connection, Optional<Extension> extension, String name, Optional<String> id, Map<String, ? extends String> configuredProperties, Iterable<? extends Dependency> dependencies) {
    return validate(new ImmutableExpressionFilterStep(stepKind, kind, action, connection, extension, name, id, configuredProperties, dependencies));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableExpressionFilterStep validate(ImmutableExpressionFilterStep instance) {
    Set<ConstraintViolation<ImmutableExpressionFilterStep>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link ExpressionFilterStep} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable ExpressionFilterStep instance
   */
  public static ExpressionFilterStep copyOf(ExpressionFilterStep instance) {
    if (instance instanceof ImmutableExpressionFilterStep) {
      return (ImmutableExpressionFilterStep) instance;
    }
    return new ExpressionFilterStep.Builder()
        .createFrom(instance)
        .build();
  }

  private Object readResolve() throws ObjectStreamException {
    return validate(this);
  }

  /**
   * Builds instances of type {@link ExpressionFilterStep ExpressionFilterStep}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  @JsonIgnoreProperties({"filterExpression"})
  public static class Builder {
    private static final long OPT_BIT_CONFIGURED_PROPERTIES = 0x1L;
    private static final long OPT_BIT_DEPENDENCIES = 0x2L;
    private long optBits;

    private String stepKind;
    private Kind kind;
    private Action<?> action;
    private Connection connection;
    private Extension extension;
    private String name;
    private String id;
    private Map<String, String> configuredProperties = new LinkedHashMap<String, String>();
    private List<Dependency> dependencies = new ArrayList<Dependency>();

    /**
     * Creates a builder for {@link ExpressionFilterStep ExpressionFilterStep} instances.
     */
    public Builder() {
      if (!(this instanceof ExpressionFilterStep.Builder)) {
        throw new UnsupportedOperationException("Use: new ExpressionFilterStep.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.integration.Step} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExpressionFilterStep.Builder createFrom(Step instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.filter.ExpressionFilterStep} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExpressionFilterStep.Builder createFrom(ExpressionFilterStep instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithKind} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExpressionFilterStep.Builder createFrom(WithKind instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithDependencies} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExpressionFilterStep.Builder createFrom(WithDependencies instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code io.syndesis.model.WithConfiguredProperties} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExpressionFilterStep.Builder createFrom(WithConfiguredProperties instance) {
      Objects.requireNonNull(instance, "instance");
      from((Object) instance);
      return (ExpressionFilterStep.Builder) this;
    }

    private void from(Object object) {
      long bits = 0;
      if (object instanceof Step) {
        Step instance = (Step) object;
        String nameValue = instance.getName();
        if (nameValue != null) {
          name(nameValue);
        }
        if ((bits & 0x1L) == 0) {
          stepKind(instance.getStepKind());
          bits |= 0x1L;
        }
        Optional<Action<?>> actionOptional = instance.getAction();
        if (actionOptional.isPresent()) {
          action(actionOptional);
        }
        Optional<Extension> extensionOptional = instance.getExtension();
        if (extensionOptional.isPresent()) {
          extension(extensionOptional);
        }
        Optional<Connection> connectionOptional = instance.getConnection();
        if (connectionOptional.isPresent()) {
          connection(connectionOptional);
        }
        if ((bits & 0x2L) == 0) {
          kind(instance.getKind());
          bits |= 0x2L;
        }
      }
      if (object instanceof ExpressionFilterStep) {
        ExpressionFilterStep instance = (ExpressionFilterStep) object;
        if ((bits & 0x1L) == 0) {
          stepKind(instance.getStepKind());
          bits |= 0x1L;
        }
        if ((bits & 0x2L) == 0) {
          kind(instance.getKind());
          bits |= 0x2L;
        }
        Optional<String> idOptional = instance.getId();
        if (idOptional.isPresent()) {
          id(idOptional);
        }
      }
      if (object instanceof WithKind) {
        WithKind instance = (WithKind) object;
        if ((bits & 0x2L) == 0) {
          kind(instance.getKind());
          bits |= 0x2L;
        }
      }
      if (object instanceof WithDependencies) {
        WithDependencies instance = (WithDependencies) object;
        addAllDependencies(instance.getDependencies());
      }
      if (object instanceof WithConfiguredProperties) {
        WithConfiguredProperties instance = (WithConfiguredProperties) object;
        putAllConfiguredProperties(instance.getConfiguredProperties());
      }
    }

    /**
     * Initializes the value for the {@link ExpressionFilterStep#getStepKind() stepKind} attribute.
     * <p><em>If not set, this attribute will have a default value as returned by the initializer of {@link ExpressionFilterStep#getStepKind() stepKind}.</em>
     * @param stepKind The value for stepKind 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("stepKind")
    public final ExpressionFilterStep.Builder stepKind(String stepKind) {
      this.stepKind = Objects.requireNonNull(stepKind, "stepKind");
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Initializes the value for the {@link ExpressionFilterStep#getKind() kind} attribute.
     * <p><em>If not set, this attribute will have a default value as returned by the initializer of {@link ExpressionFilterStep#getKind() kind}.</em>
     * @param kind The value for kind 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("kind")
    public final ExpressionFilterStep.Builder kind(Kind kind) {
      this.kind = Objects.requireNonNull(kind, "kind");
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExpressionFilterStep#getAction() action} to action.
     * @param action The value for action
     * @return {@code this} builder for chained invocation
     */
    public final ExpressionFilterStep.Builder action(Action<?> action) {
      this.action = Objects.requireNonNull(action, "action");
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExpressionFilterStep#getAction() action} to action.
     * @param action The value for action
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("action")
    public final ExpressionFilterStep.Builder action(Optional<? extends Action<?>> action) {
      this.action = action.orElse(null);
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExpressionFilterStep#getConnection() connection} to connection.
     * @param connection The value for connection
     * @return {@code this} builder for chained invocation
     */
    public final ExpressionFilterStep.Builder connection(Connection connection) {
      this.connection = Objects.requireNonNull(connection, "connection");
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExpressionFilterStep#getConnection() connection} to connection.
     * @param connection The value for connection
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("connection")
    public final ExpressionFilterStep.Builder connection(Optional<? extends Connection> connection) {
      this.connection = connection.orElse(null);
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExpressionFilterStep#getExtension() extension} to extension.
     * @param extension The value for extension
     * @return {@code this} builder for chained invocation
     */
    public final ExpressionFilterStep.Builder extension(Extension extension) {
      this.extension = Objects.requireNonNull(extension, "extension");
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExpressionFilterStep#getExtension() extension} to extension.
     * @param extension The value for extension
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("extension")
    public final ExpressionFilterStep.Builder extension(Optional<? extends Extension> extension) {
      this.extension = extension.orElse(null);
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Initializes the value for the {@link ExpressionFilterStep#getName() name} attribute.
     * @param name The value for name (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("name")
    public final ExpressionFilterStep.Builder name(String name) {
      this.name = name;
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExpressionFilterStep#getId() id} to id.
     * @param id The value for id
     * @return {@code this} builder for chained invocation
     */
    public final ExpressionFilterStep.Builder id(String id) {
      this.id = Objects.requireNonNull(id, "id");
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Initializes the optional value {@link ExpressionFilterStep#getId() id} to id.
     * @param id The value for id
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("id")
    public final ExpressionFilterStep.Builder id(Optional<String> id) {
      this.id = id.orElse(null);
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Put one entry to the {@link ExpressionFilterStep#getConfiguredProperties() configuredProperties} map.
     * @param key The key in the configuredProperties map
     * @param value The associated value in the configuredProperties map
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExpressionFilterStep.Builder putConfiguredProperty(String key, String value) {
      this.configuredProperties.put(
          Objects.requireNonNull(key, "configuredProperties key"),
          Objects.requireNonNull(value, "configuredProperties value"));
      optBits |= OPT_BIT_CONFIGURED_PROPERTIES;
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Put one entry to the {@link ExpressionFilterStep#getConfiguredProperties() configuredProperties} map. Nulls are not permitted
     * @param entry The key and value entry
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExpressionFilterStep.Builder putConfiguredProperty(Map.Entry<String, ? extends String> entry) {
      String k = entry.getKey();
      String v = entry.getValue();
      this.configuredProperties.put(
          Objects.requireNonNull(k, "configuredProperties key"),
          Objects.requireNonNull(v, "configuredProperties value"));
      optBits |= OPT_BIT_CONFIGURED_PROPERTIES;
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Sets or replaces all mappings from the specified map as entries for the {@link ExpressionFilterStep#getConfiguredProperties() configuredProperties} map. Nulls are not permitted
     * @param configuredProperties The entries that will be added to the configuredProperties map
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("configuredProperties")
    public final ExpressionFilterStep.Builder configuredProperties(Map<String, ? extends String> configuredProperties) {
      this.configuredProperties.clear();
      optBits |= OPT_BIT_CONFIGURED_PROPERTIES;
      return putAllConfiguredProperties(configuredProperties);
    }

    /**
     * Put all mappings from the specified map as entries to {@link ExpressionFilterStep#getConfiguredProperties() configuredProperties} map. Nulls are not permitted
     * @param configuredProperties The entries that will be added to the configuredProperties map
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExpressionFilterStep.Builder putAllConfiguredProperties(Map<String, ? extends String> configuredProperties) {
      for (Map.Entry<String, ? extends String> entry : configuredProperties.entrySet()) {
        String k = entry.getKey();
        String v = entry.getValue();
        this.configuredProperties.put(
            Objects.requireNonNull(k, "configuredProperties key"),
            Objects.requireNonNull(v, "configuredProperties value"));
      }
      optBits |= OPT_BIT_CONFIGURED_PROPERTIES;
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Adds one element to {@link ExpressionFilterStep#getDependencies() dependencies} list.
     * @param element A dependencies element
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExpressionFilterStep.Builder addDependency(Dependency element) {
      this.dependencies.add(Objects.requireNonNull(element, "dependencies element"));
      optBits |= OPT_BIT_DEPENDENCIES;
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Adds elements to {@link ExpressionFilterStep#getDependencies() dependencies} list.
     * @param elements An array of dependencies elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExpressionFilterStep.Builder addDependency(Dependency... elements) {
      for (Dependency element : elements) {
        this.dependencies.add(Objects.requireNonNull(element, "dependencies element"));
      }
      optBits |= OPT_BIT_DEPENDENCIES;
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Sets or replaces all elements for {@link ExpressionFilterStep#getDependencies() dependencies} list.
     * @param elements An iterable of dependencies elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("dependencies")
    public final ExpressionFilterStep.Builder dependencies(Iterable<? extends Dependency> elements) {
      this.dependencies.clear();
      return addAllDependencies(elements);
    }

    /**
     * Adds elements to {@link ExpressionFilterStep#getDependencies() dependencies} list.
     * @param elements An iterable of dependencies elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final ExpressionFilterStep.Builder addAllDependencies(Iterable<? extends Dependency> elements) {
      for (Dependency element : elements) {
        this.dependencies.add(Objects.requireNonNull(element, "dependencies element"));
      }
      optBits |= OPT_BIT_DEPENDENCIES;
      return (ExpressionFilterStep.Builder) this;
    }

    /**
     * Builds a new {@link ExpressionFilterStep ExpressionFilterStep}.
     * @return An immutable instance of ExpressionFilterStep
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ExpressionFilterStep build() {
      return ImmutableExpressionFilterStep.validate(new ImmutableExpressionFilterStep(this));
    }

    private boolean configuredPropertiesIsSet() {
      return (optBits & OPT_BIT_CONFIGURED_PROPERTIES) != 0;
    }

    private boolean dependenciesIsSet() {
      return (optBits & OPT_BIT_DEPENDENCIES) != 0;
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<T>();
    } else {
      list = new ArrayList<T>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<T>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }

  private static <K, V> Map<K, V> createUnmodifiableMap(boolean checkNulls, boolean skipNulls, Map<? extends K, ? extends V> map) {
    switch (map.size()) {
    case 0: return Collections.emptyMap();
    case 1: {
      Map.Entry<? extends K, ? extends V> e = map.entrySet().iterator().next();
      K k = e.getKey();
      V v = e.getValue();
      if (checkNulls) {
        Objects.requireNonNull(k, "key");
        Objects.requireNonNull(v, "value");
      }
      if (skipNulls && (k == null || v == null)) {
        return Collections.emptyMap();
      }
      return Collections.singletonMap(k, v);
    }
    default: {
      Map<K, V> linkedMap = new LinkedHashMap<K, V>(map.size());
      if (skipNulls || checkNulls) {
        for (Map.Entry<? extends K, ? extends V> e : map.entrySet()) {
          K k = e.getKey();
          V v = e.getValue();
          if (skipNulls) {
            if (k == null || v == null) continue;
          } else if (checkNulls) {
            Objects.requireNonNull(k, "key");
            Objects.requireNonNull(v, "value");
          }
          linkedMap.put(k, v);
        }
      } else {
        linkedMap.putAll(map);
      }
      return Collections.unmodifiableMap(linkedMap);
    }
    }
  }
}
