/*
 * Copyright (C) 2016 Red Hat, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.syndesis.model.icon;

import com.fasterxml.jackson.annotation.JsonProperty;
import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import javax.annotation.Generated;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.Validation;
import javax.validation.Validator;

/**
 * Immutable implementation of {@link Icon}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code new Icon.Builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableIcon.of()}.
 */
@SuppressWarnings({"all"})
@SuppressFBWarnings
@Generated({"Immutables.generator", "Icon"})
final class ImmutableIcon implements Icon {
  private final String mediaType;
  private final String id;

  private ImmutableIcon(String mediaType, Optional<String> id) {
    this.mediaType = mediaType;
    this.id = id.orElse(null);
  }

  private ImmutableIcon(ImmutableIcon original, String mediaType, String id) {
    this.mediaType = mediaType;
    this.id = id;
  }

  /**
   * @return The value of the {@code mediaType} attribute
   */
  @JsonProperty("mediaType")
  @Override
  public String getMediaType() {
    return mediaType;
  }

  /**
   * @return The value of the {@code id} attribute
   */
  @JsonProperty("id")
  @Override
  public Optional<String> getId() {
    return Optional.ofNullable(id);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Icon#getMediaType() mediaType} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for mediaType (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableIcon withMediaType(String value) {
    if (Objects.equals(this.mediaType, value)) return this;
    return validate(new ImmutableIcon(this, value, this.id));
  }

  /**
   * Copy the current immutable object by setting a <i>present</i> value for the optional {@link Icon#getId() id} attribute.
   * @param value The value for id
   * @return A modified copy of {@code this} object
   */
  public final ImmutableIcon withId(String value) {
    String newValue = Objects.requireNonNull(value, "id");
    if (Objects.equals(this.id, newValue)) return this;
    return validate(new ImmutableIcon(this, this.mediaType, newValue));
  }

  /**
   * Copy the current immutable object by setting an optional value for the {@link Icon#getId() id} attribute.
   * An equality check is used on inner nullable value to prevent copying of the same value by returning {@code this}.
   * @param optional A value for id
   * @return A modified copy of {@code this} object
   */
  public final ImmutableIcon withId(Optional<String> optional) {
    String value = optional.orElse(null);
    if (Objects.equals(this.id, value)) return this;
    return validate(new ImmutableIcon(this, this.mediaType, value));
  }

  /**
   * This instance is equal to all instances of {@code ImmutableIcon} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableIcon
        && equalTo((ImmutableIcon) another);
  }

  private boolean equalTo(ImmutableIcon another) {
    return Objects.equals(mediaType, another.mediaType)
        && Objects.equals(id, another.id);
  }

  /**
   * Computes a hash code from attributes: {@code mediaType}, {@code id}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + Objects.hashCode(mediaType);
    h += (h << 5) + Objects.hashCode(id);
    return h;
  }

  /**
   * Prints the immutable value {@code Icon} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    StringBuilder builder = new StringBuilder("Icon{");
    if (mediaType != null) {
      builder.append("mediaType=").append(mediaType);
    }
    if (id != null) {
      if (builder.length() > 5) builder.append(", ");
      builder.append("id=").append(id);
    }
    return builder.append("}").toString();
  }

  /**
   * Construct a new immutable {@code Icon} instance.
   * @param mediaType The value for the {@code mediaType} attribute
   * @param id The value for the {@code id} attribute
   * @return An immutable Icon instance
   */
  public static Icon of(String mediaType, Optional<String> id) {
    return validate(new ImmutableIcon(mediaType, id));
  }

  private static final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();

  private static ImmutableIcon validate(ImmutableIcon instance) {
    Set<ConstraintViolation<ImmutableIcon>> constraintViolations = validator.validate(instance);
    if (!constraintViolations.isEmpty()) {
      throw new ConstraintViolationException(constraintViolations);
    }
    return instance;
  }

  /**
   * Creates an immutable copy of a {@link Icon} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Icon instance
   */
  public static Icon copyOf(Icon instance) {
    if (instance instanceof ImmutableIcon) {
      return (ImmutableIcon) instance;
    }
    return new Icon.Builder()
        .createFrom(instance)
        .build();
  }

  /**
   * Builds instances of type {@link Icon Icon}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static class Builder {
    private String mediaType;
    private String id;

    /**
     * Creates a builder for {@link Icon Icon} instances.
     */
    public Builder() {
      if (!(this instanceof Icon.Builder)) {
        throw new UnsupportedOperationException("Use: new Icon.Builder()");
      }
    }

    /**
     * Fill a builder with attribute values from the provided {@code Icon} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Icon.Builder createFrom(Icon instance) {
      Objects.requireNonNull(instance, "instance");
      String mediaTypeValue = instance.getMediaType();
      if (mediaTypeValue != null) {
        mediaType(mediaTypeValue);
      }
      Optional<String> idOptional = instance.getId();
      if (idOptional.isPresent()) {
        id(idOptional);
      }
      return (Icon.Builder) this;
    }

    /**
     * Initializes the value for the {@link Icon#getMediaType() mediaType} attribute.
     * @param mediaType The value for mediaType (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("mediaType")
    public final Icon.Builder mediaType(String mediaType) {
      this.mediaType = mediaType;
      return (Icon.Builder) this;
    }

    /**
     * Initializes the optional value {@link Icon#getId() id} to id.
     * @param id The value for id
     * @return {@code this} builder for chained invocation
     */
    public final Icon.Builder id(String id) {
      this.id = Objects.requireNonNull(id, "id");
      return (Icon.Builder) this;
    }

    /**
     * Initializes the optional value {@link Icon#getId() id} to id.
     * @param id The value for id
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("id")
    public final Icon.Builder id(Optional<String> id) {
      this.id = id.orElse(null);
      return (Icon.Builder) this;
    }

    /**
     * Builds a new {@link Icon Icon}.
     * @return An immutable instance of Icon
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public Icon build() {
      return ImmutableIcon.validate(new ImmutableIcon(null, mediaType, id));
    }
  }
}
